/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.auxml;

import java.io.File;
import java.io.FileWriter;
import java.io.FileFilter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.Map;

/**
 *
 * @author Martin.Schovanek@sun.com
 */
public class AuManager {

    public static final String MODULES_SETUP = "modules.setup";
    public static final String NBMS_PATH = "nbms";
    public static final String REMOVED_PATH = "removed";
    public static final String SETUP_PATH = "catalogs.setup";
    public static final String TEMPLATES_PATH = "templates";
    public static final String STAGING = "_staging_";
    public static final String LIVE = "modules";
    public static final String PUBLISH = "publish";
    public static final String STAGE = "stage";
    public static final String REMOVE = "remove";
    private File ucRoot;
    private File catalogsSetup;
    private File nbms;
    private File templates;
    private File removed;

    private AuManager(File ucRoot) {
        this.ucRoot = ucRoot;
        catalogsSetup = new File(ucRoot, SETUP_PATH);
        nbms = new File(ucRoot, NBMS_PATH);
        removed = new File(nbms, REMOVED_PATH);
        templates = new File(ucRoot, TEMPLATES_PATH);
    }

    public static AuManager getInstance(String ucRoot) {
        return new AuManager(new File(ucRoot));
    }

    public UpdateCenter getCatalog(String version, String type) throws IOException {
        return new UpdateCenter(ucRoot.getAbsolutePath(), version, type);
    }

    public Version getVersion(String version) {
        return new Version(version, ucRoot);
    }

    public List<Version> getVersions() {
        List<Version> versions;
        File[] files = catalogsSetup.listFiles(new FileFilter() {

            public boolean accept(File file) {
                return new File(file, MODULES_SETUP).exists();
            }
        });
        if (files == null || files.length == 0) {
            versions = Collections.EMPTY_LIST;
        } else {
            versions = new ArrayList<Version>(files.length);
            for (File file : files) {
                versions.add(new Version(file.getName(), ucRoot));
            }
        }
        return versions;
    }

    public void createVersion(String name) {
        File versionSetupDir = mkdir(catalogsSetup, name);
        String str =
                "##################################\n" +
                "# Setup for Netbeans " + name + "\n" +
                "##################################\n";
        FileOutputStream fos;
        try {
            fos = new FileOutputStream(new File(versionSetupDir, MODULES_SETUP));
            try {
                fos.write(str.getBytes());
            } finally {
                fos.close();
            }
        } catch (IOException ex) {
            Logger.getLogger(AuManager.class.getName()).log(Level.SEVERE, null, ex);
        }

        mkdir(nbms, name);
    }

    public void createUpdatCenter(String version, String type, String template) throws IOException {
        File tempFile = getTemplate(template);
        if (!tempFile.canRead()) {
            throw new IllegalStateException("Cannot read template file: " + tempFile.getAbsolutePath());
        }
        Version ver = getVersion(version);
        // create modules' dirs
        File ucDir = new File(ver.getModulesDir(), type);
        if (!ucDir.mkdir()) {
            throw new IllegalStateException(type + " - Update Center alredy exists.");
        }
        new File(ucDir, "modules").mkdir();
        new File(ucDir, "_staging_").mkdir();
        // create catalogs.setup file
        Properties prop = new Properties();
        prop.setProperty("catalog_template", TEMPLATES_PATH + "/" + template);
        File setupFile = new File(ver.getSetupDir(), "setup." + version + "." + type);
        FileOutputStream fos = new FileOutputStream(setupFile);
        try {
            prop.store(fos, null);
        } finally {
            fos.close();
        }
        UpdateCenter uc = getCatalog(version, type);
        uc.generateCatalog();
    }

    public File getTemplate(String name) {
        return new File(templates, name);
    }

    public File[] getTemplates() {
        File[] files = templates.listFiles(new FileFilter() {

            public boolean accept(File file) {
                return file.getName().endsWith(".xsl") && file.isFile();
            }
        });
        return files;
    }

    public void stageModule(String version, String type, String folder, File nbm) throws IOException {
        // validate nbm (move into separate function) TODO
        ModuleGroup mg = new ModuleGroup("test_module_group", null);
        Module module = new Module(new NBMfile(nbm), mg);
        String license = module.getLicense();
        if (license == null || license.trim().equals("")) {
            throw new IllegalArgumentException("Missing License in: " + nbm.getAbsolutePath());
        }
        // sign the nbm TODO
        // move the nbm into _staging_ folder
        UpdateCenter uc = getCatalog(version, type);
        
        File staging_dir = new File(uc.getCatalogDir(), STAGING);
        if ( folder!=null && folder.length() > 0) {
            File folder_dir = new File(staging_dir, folder);
            if (!folder_dir.exists()) {
                mkdir(staging_dir, folder);
            }
            staging_dir = folder_dir;
        }
        
        File destFile = new File(staging_dir, nbm.getName());
        //copyFile(nbm, destFile);
        mvFile(nbm, destFile);
        // generate catalog
        uc.generateCatalog();
    }
    
    
    public void applyChanges(String version, String type, Map<String, String[]> params) throws IOException {
        
        UpdateCenter uc = getCatalog(version, type);
        File uc_dir = uc.getCatalogDir();
        
        String[] module, action;
        String nbm_name;
        File nbm_file, nbm_path, nbm_path_parent;
        int i=0;
        while ( (module = params.get("stg_module" + i)) != null ) {
            action = params.get("stg_module_act" + i);
            
            if ( action[0].equals(REMOVE)) {        
               nbm_file = new File(uc_dir, module[0]);
               nbm_name = nbm_file.getName();
               mvFile(nbm_file, new File(removed, nbm_name));
            }
            if ( action[0].equals(PUBLISH)) {
               nbm_file = new File(uc_dir, module[0]);
               nbm_name = nbm_file.getAbsolutePath();
               nbm_name = nbm_name.replaceFirst(STAGING, LIVE);
               
               nbm_path = new File(nbm_name);
               nbm_path_parent = nbm_path.getParentFile();
               mkdir(nbm_path_parent);
               
               mvFile(nbm_file, nbm_path);
            }
            i++;
        }
        
        i=0;
        while ( (module = params.get("pub_module" + i)) != null ) {
            action = params.get("pub_module_act" + i);
            
            if ( action[0].equals(REMOVE)) {        
               nbm_file = new File(uc_dir, module[0]);
               nbm_name = nbm_file.getName();
               mvFile(nbm_file, new File(removed, nbm_name));
            }
            if ( action[0].equals(STAGE)) {
               nbm_file = new File(uc_dir, module[0]);
               nbm_name = nbm_file.getAbsolutePath();
               nbm_name = nbm_name.replaceFirst(LIVE, STAGING);
               
               nbm_path = new File(nbm_name);
               nbm_path_parent = nbm_path.getParentFile();
               mkdir(nbm_path_parent);
               
               mvFile(nbm_file, nbm_path);
            }

        }
        uc.generateCatalog();
        i++;
    }
    

    private void mvFile(File src, File dest) {
        if(!dest.getParentFile().exists()) {
            dest.getParentFile().mkdirs();
        }
        if (!src.renameTo(dest)) {
            throw new IllegalStateException("Cannot move file: " + src.getAbsolutePath() + "\n into: " + dest.getAbsolutePath());
        }
    }

    // XXX method not used!!!
    private void copyFile(File src, File dest) {
        try {
            FileChannel inc = new FileInputStream(src).getChannel();
            try {
                FileChannel outc = new FileOutputStream(dest).getChannel();
                try {
                    inc.transferTo(0, inc.size(), outc);
                } finally {
                    outc.close();
                }
            } finally {
                inc.close();
            }
        } catch (IOException ex) {
            throw new IllegalStateException("Cannot copy file: " + src.getAbsolutePath() + "\n into: " + dest.getAbsolutePath(), ex);
        }
    }

    private File mkdir(File parent, String name) {
        File dir = new File(parent, name);
        if (dir.exists())
            return dir;
        
        if (!dir.mkdirs()) {
            throw new IllegalStateException("Cannot create folder: " + dir.getAbsolutePath());
        }
        return dir;
    }
    
    private File mkdir(File dir) {
        if (dir.exists())
            return dir;
        
        if (!dir.mkdirs()) {
            throw new IllegalStateException("Cannot create folder: " + dir.getAbsolutePath());
        }
        return dir;
    }
}
