/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.auxml;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.junit.Assert;
import org.w3c.dom.Document;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import static org.junit.Assert.*;
/**
 *
 * @author ms113234
 */
public class Utils {
    private static final Logger LOG = Logger.getLogger(Utils.class.getName());

    public static void assertEquals(File ref, File golden) {
        Assert.assertEquals(fileToString(ref), fileToString(golden));
    }

    public static String fileToString(File file) {
        StringBuilder sb = new StringBuilder(256 * 1024);
        byte[] buff = new byte[64 * 1024];
        try {
            FileInputStream fis = new FileInputStream(file);
            try {
                int n = fis.read(buff);
                while (n > 0) {
                    sb.append(new String(buff,0,n));
                    n = fis.read(buff);
                }
            } catch (IOException ex) {
                ex.printStackTrace();
                Assert.fail("Cannot read: " + file);
            } finally {
                fis.close();
            }
        } catch (IOException ioe) {
            ioe.printStackTrace();
            Assert.fail("Cannot read: " + file);
        }
        return sb.toString();
    }

    public static File goldenFile(String version, String type, String fileName) {
        File file = null;
        try {
            StringBuilder sb = new StringBuilder("goldenfiles/");
            sb.append("/").append(version).append("/").append(type);
            sb.append("/").append(fileName);
            file = new File(Utils.class.getResource(sb.toString()).toURI());
        } catch (URISyntaxException ex) {
            LOG.log(Level.SEVERE, null, ex);
        }
        return file;
    }
    
    public static void verifyCatalogs(UpdateCenter uc) throws IOException {
        AssertionError error = null;
        for (boolean stage : new boolean[]{false, true}) {
            List<String> languages = new ArrayList<String>(uc.getLanguages());
            languages.add(null); // add default catalog
            for (String lang : languages) {
                // read catalog
                File catFile = uc.getCatalogFile(stage, lang);
                String catData = fileToString(catFile);
                String name = catFile.getName();
                // copy catalog betwen goldenfiles for easily diffing
                File copy = new File(goldenFile(uc.getVersion(),uc.getType(),""), name.replaceFirst(".xml", "-ref.xml"));
                System.err.println("Writing catalog copy into: " + copy.getAbsolutePath());
                FileOutputStream fos = new FileOutputStream(copy);
                fos.write(catData.getBytes());
                fos.close();
                // clean timestamp and compare DOM with the golden file
                Document catDoc;
                Document goldDoc;
                File goldFile = goldenFile(uc.getVersion(),uc.getType(),name);
                catDoc = parse(catFile);
                catDoc.getDocumentElement().getAttributeNode("timestamp").setValue("");
                goldDoc = parse(goldFile);
                goldDoc.getDocumentElement().getAttributeNode("timestamp").setValue("");
                // postpone possible fail to the end of the test
                try {
                    assertTrue(DOMCompare.compareDocuments(catDoc, goldDoc));
                } catch (AssertionError ae) {
                    System.err.println("ERROR: compariton failed in: " + catFile.getAbsolutePath());
                    if (error == null) {
                        error = ae;
                    }
                }
            }
        }
        if (error != null) {
            throw error;
        }
    }

    /**
     * Create from factory a DocumentBuilder and let it create a org.w3c.dom.Document.
     * This method takes InputSource. After successful finish the document tree is returned.
     * (Implemtation comes from org.openide.xml.XMLUtil);
     *
     * @param input a parser input (for URL users use: <code>new InputSource(url.toExternalForm())</code>
     * @param validate if true validating parser is used
     * @param namespaceAware if true DOM is created by namespace aware parser
     * @param errorHandler a error handler to notify about exception or <code>null</code>
     * @param entityResolver SAX entity resolver or <code>null</code>; see class Javadoc for hints
     *
     * @throws IOException if an I/O problem during parsing occurs
     * @throws SAXException is thrown if a parser error occurs
     *
     * @return document representing given input, or null if a parsing error occurs
     */
    public static Document parse(
        InputSource input, boolean validate, boolean namespaceAware, ErrorHandler errorHandler,
        EntityResolver entityResolver
    ) throws IOException, SAXException {
        DocumentBuilder builder = null;
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setValidating(validate);
        factory.setNamespaceAware(namespaceAware);
        try {
            builder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException ex) {
            throw new SAXException("Cannot create parser satisfying configuration parameters", ex); //NOI18N
        }
        if (errorHandler != null) {
            builder.setErrorHandler(errorHandler);
        }
        if (entityResolver != null) {
            builder.setEntityResolver(entityResolver);
        }
        return builder.parse(input);
    }
    
    public static Document parse(File input) {
        Document doc = null;
        try {
            FileInputStream fis = new FileInputStream(input);
            try {
                doc = parse(new InputSource(fis), false, true, null, null);
            } catch (Exception ex) {
                LOG.log(Level.SEVERE, null, ex);
            } finally {
                fis.close();
            }
        } catch (IOException ex) {
                LOG.log(Level.SEVERE, null, ex);
        }
        return doc;
    }
}
