/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */

package org.netbeans.modules.coverage.config;

import java.util.Properties;
import org.netbeans.api.project.Project;
import org.netbeans.modules.coverage.exceptions.PluginNotConfiguredException;
import org.netbeans.modules.coverage.instrumentation.XmlBuilder;
import org.netbeans.modules.coverage.utils.ModuleActivityModes;
import org.openide.filesystems.FileObject;

/**
 * The class holds code coverage plugin configuration data for all projects 
 * and provides access to this data. Config instance that holds data can be obtained for 
 * a specific project. The class was planned as the single configuartion access point
 * 
 * TODO: Add configuration caching later
 * @author mikk@netbeans.org
 */
public class Config {
    
    /** Creates a new instance of Config */
    private Config() {
    }
    private Properties properties = null;
    
    /**
     * Creates and initializes a config instance for the specified project
     * @param project The project to get Config for
     * @throws org.netbeans.modules.coverage.exceptions.PluginNotConfiguredException Throws this exception if plugin data is not available
     * @return returns Config class instance for the specified project
     */
    public static Config getInstanceForProject(Project project) throws PluginNotConfiguredException{
        Config instance = new Config();
        Properties props = ProjectPropertiesReader.loadProjectProperties(project);
        
        if (!arePropertiesConfigured(props))
            throw new PluginNotConfiguredException();
        
        instance.properties = props;
        return instance;
    }
    
    /**
     * Stores project properties to disk
     * @param project The project to store properties for
     * @param props Properties to be saved
     */
    public static void storeProjectProperties(Project project, Properties props){
        ProjectPropertiesReader.storeProjectPropertiesToFile(project, props);
    }
    
    
    /**
     * Returns properties for the project
     * @return Returns proeprties for the project
     */
    public Properties getPluginProperties(){
        return properties;
    }
    
    /**
     * Returns versioning mode
     * @return versioning mode
     */
    public String getVersioningMode() {
        return properties.getProperty("coverage.versioning");
    }
    
    /**
     * Says if the module is active based on the properties supplied
     * @param props The props to analyze
     * @return Returns True if the module is active
     */
    public static boolean isModuleActive(Properties props){
        if (props.containsKey("coverage.activity")
        && props.getProperty("coverage.activity").equals(ModuleActivityModes.ON)){
            return true;
        }else{
            return false;
        }
    }
    
    /**
     * Says if the module is active
     * @return true, if the module is active
     */
    public boolean isModuleActive(){
        return isModuleActive(properties);
    }
    
    /**
     *The method tells if all required properties are set for a project
     **/
    private static boolean arePropertiesConfigured(Properties props){
        if (props.containsKey("coverage.activity")
        && props.containsKey("coverage.coverageFile")
        && props.containsKey("coverage.templateFile")){
            return true;
        }else{
            return false;
        }
    }

    /**
     * Project type constant for NB Module
     */
    public static final String NB_MODULE = "module";
    /**
     * Project type constant for java ptojects
     */
    public static final String JAVA_PROJECT = "java";
    /**
     * Project type constant for other project types
     */
    public static final String UNKNOWN = "unknown";
    /**
     * Returns project type by project
     * @param project project to get type for
     * @return Project type constant
     */
    public static String getProjectType(Project project){
        String projType = UNKNOWN;
        String projectClass = project.getClass().getName();
        if (projectClass.equals("org.netbeans.modules.apisupport.project.NbModuleProject")){
            projType = NB_MODULE;
        }
        if (projectClass.equals("org.netbeans.modules.java.j2seproject.J2SEProject")){
            projType = JAVA_PROJECT;
        }
        return projType;
     }

    /**
     * Says if the project is a java project
     * @param project Project to check
     * @return Check result
     */
    public static boolean isJavaProject(Project project){
        return getProjectType(project).equals(JAVA_PROJECT);
    }
    
    /**
     * Checks if the project is a NB module project
     * @param project Project to check
     * @return Check result
     */
    public static boolean isNbModuleProject(Project project){
        return getProjectType(project).equals(NB_MODULE);
    }
    
    public static boolean isSupportedProject(Project project){
        return (isNbModuleProject(project) || isJavaProject(project));
    }
    
}
