package org.netbeans.jsstubsgenerator;

import java.io.File;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.netbeans.jsstubsgenerator.core.NoSpecCoreGenerator;
import org.netbeans.jsstubsgenerator.core.ecma.EcmaCoreGenerator;
import org.netbeans.jsstubsgenerator.core.mdn.ArgumentsGenerator;
import org.netbeans.jsstubsgenerator.core.mdn.MdnGenerator;
import org.netbeans.jsstubsgenerator.core.mdn.XMLHttpRequestGenerator;
import org.netbeans.jsstubsgenerator.dom.DomGenerator;
import org.netbeans.jsstubsgenerator.dom.HistoryGenerator;
import org.netbeans.jsstubsgenerator.dom.LocationGenerator;
import org.netbeans.jsstubsgenerator.dom.StorageGenerator;
import org.netbeans.jsstubsgenerator.dom.WindowGenerator;
import org.netbeans.jsstubsgenerator.dom.ConsoleInstanceGenerator;
import org.netbeans.jsstubsgenerator.dom.CryptoInstanceGenerator;
import org.netbeans.jsstubsgenerator.dom.DocumentInstanceGenerator;
import org.netbeans.jsstubsgenerator.dom.ElementTraversalGenerator;
import org.netbeans.jsstubsgenerator.dom.NavigatorInstanceGenerator;
import org.netbeans.jsstubsgenerator.dom.ScreenInstanceGenerator;
import org.netbeans.jsstubsgenerator.dom.WindowInstanceGenerator;
import org.netbeans.jsstubsgenerator.stub.StubObject;

/**
 * Generates JavaScript stubs for purpose of code completion and its documentation.
 *
 * TODO - parse since values TODO - compatibility?
 *
 * @author Martin Fousek <marfous@netbeans.org>
 */
public class JsStubsGenerator {

    /** Whether to generate legal stubs for distribution or full for UC. */
    public static final boolean GENERATE_FULL = false;

    /** Whether to pack all stubs into bundles according to licenses. */
    public static final boolean PACK_BUNDLES = true;

    /** Whether to generate into output code for ReturnHolders creation - no doc generated then. */
    public static final boolean GENERATE_RETURN_HOLDERS = false;

    private static final Logger LOGGER = Logger.getLogger(JsStubsGenerator.class.getName());
    private static final String MDN_SUBDIR = "developer.mozilla.org" + File.separator + "en";
    private static final String ECMA_SUBDIR = "core";
    private static final String W3C_SUBDIR = "w3.org";
    private static File mdnInput;
    private static File ecmaInput;
    private static File w3cInput;
    private static File output;
    private static Map<String, StubObject> allStubs = new HashMap<String, StubObject>();

    public static void main(String[] args) {
        args = new String[3];
        args[0] = "/home/marfous/Work/NetBeans/Javascript2/jsstubs2";
        args[1] = "/home/marfous/Work/NetBeans/Javascript2/jsstubs2/allstubs";

        if (args.length < 2) {
            LOGGER.log(Level.SEVERE, "Usage: {0} <inputdir> <outputdir>", JsStubsGenerator.class.getName());
            System.exit(-1);
        }

        initInputArguments(args);
        generateDocs();
        if (PACK_BUNDLES) {
            packBundles();
        }
    }

    private static void initInputArguments(String[] args) {
        File inputDir = new File(args[0]);
        if (!inputDir.exists()) {
            LOGGER.log(Level.WARNING, "{0} does not exist", inputDir.getPath());
            System.exit(1);
        }

        mdnInput = new File(inputDir, MDN_SUBDIR);
        if (!mdnInput.exists()) {
            LOGGER.log(Level.WARNING, "Can''t find {0}", mdnInput.getPath());
            LOGGER.log(Level.WARNING, "Is this a valid input directory? See the javadoc header for\ninstructions on how to generate it.\n");
            System.exit(2);
        }

        w3cInput = new File(inputDir, W3C_SUBDIR);
        if (!w3cInput.exists()) {
            LOGGER.log(Level.WARNING, "Can''t find {0}", w3cInput.getPath());
            LOGGER.log(Level.WARNING, "Is this a valid input directory? See the javadoc header for\ninstructions on how to generate it.\n");
            System.exit(2);
        }

        ecmaInput = new File(inputDir, ECMA_SUBDIR);
        if (!ecmaInput.exists()) {
            LOGGER.log(Level.WARNING, "Can''t find {0}", w3cInput.getPath());
            LOGGER.log(Level.WARNING, "Is this a valid input directory? See the javadoc header for\ninstructions on how to generate it.\n");
            System.exit(2);
        }

        output = new File(args[1]);
        if (output.exists() && output.isFile()) {
            LOGGER.log(Level.WARNING, "Output directory must be a directory!");
            System.exit(3);
        }
    }

    private static void generateDocs() {
        // the order of the generator calls is important since in case of docs pruning we must have place where to prune
        try {
            // JavaScript core APIs - from MDN
            allStubs.putAll(MdnGenerator.parseCoreDocs(mdnInput));

            // should it create full or pruned doc
            if (!GENERATE_FULL) {
                // JavaScript core APIs - from ecma-international.org: update and prune if needed
                allStubs = EcmaCoreGenerator.getUpdatedStubs(ecmaInput, allStubs);
            }
            // complete CORE stubs for non-spec defined properties and methods
            allStubs = NoSpecCoreGenerator.updateStubs(allStubs);

            // XMLHttpRequest is special case which is pruned on base of GENERATE_FULL flag
            XMLHttpRequestGenerator.generateXMLHttpRequest(allStubs);

            // Manually created files
            ArgumentsGenerator.generateArguments(allStubs);

            // DOM APIs
            // MDN ones
            allStubs.putAll(MdnGenerator.parseDomDocs(mdnInput));
            // W3C ones
            ElementTraversalGenerator.generate(allStubs);
            DomGenerator.parseDomDocs(allStubs, w3cInput);
            LocationGenerator.generate(allStubs);
            WindowGenerator.generate(allStubs);
            WindowInstanceGenerator.generate(allStubs);
            HistoryGenerator.generate(allStubs);
            CryptoInstanceGenerator.generate(allStubs);
            ConsoleInstanceGenerator.generate(allStubs);
            NavigatorInstanceGenerator.generate(allStubs);
            ScreenInstanceGenerator.generate(allStubs);
            StorageGenerator.generate(allStubs);
            DocumentInstanceGenerator.generate(allStubs);

            if (!GENERATE_RETURN_HOLDERS) {
                generateCode();
            } else {
                generateReturnHolders();
            }

//            dump();
        } catch (IOException ioe) {
            LOGGER.log(Level.SEVERE, null, ioe);
        }
    }

    private static void generateCode() throws IOException {
        // create dirs for output
        output.mkdirs();
        for (StubObject stub : allStubs.values()) {
            File outputFile = new File(output, "stub_" + stub.getOrigin() + "_" + stub.getName() + ".js");
            FileWriter writer = new FileWriter(outputFile);
            String doc = stub.toDoc(false);
            writer.write(doc);
            writer.close();

            Parser sourceParser = new Parser();
            sourceParser.parse(doc, outputFile.getPath(), 0);
            assert sourceParser.wasSourceOk() : outputFile;

            // TODO - check for validity
//            checkValidJs(outputFile, stub);
        }
    }

    private static void dump() {
        for (Map.Entry<String, StubObject> entry : allStubs.entrySet()) {
            StubObject stub = entry.getValue();
            System.out.println(stub.dump());
        }
    }

    private static void generateReturnHolders() {
        for (StubObject stub : allStubs.values()) {
            System.err.println(stub.toReturnHolder());
        }
    }

    private static void packBundles() {
        // prepare folders
        File coreStubs = new File(output, "core/jsstubs");
        coreStubs.mkdirs();
        File domStubs = new File(output, "dom/jsstubs");
        domStubs.mkdirs();
        File restStubs = new File(output, "rest/jsstubs");
        restStubs.mkdirs();

        // move files to prepared folders
        separateDocFiles(coreStubs, domStubs, restStubs);

        // zip all folders to bundles
        String suffix = GENERATE_FULL ? "-doc" : "";
        Zipper zipper = new Zipper(coreStubs.getParent(), output.getParent() + "/js-corestubs" + suffix + ".zip");
        zipper.process();
        zipper = new Zipper(domStubs.getParent(), output.getParent() + "/js-domstubs" + suffix + ".zip");
        zipper.process();
        zipper = new Zipper(restStubs.getParent(), output.getParent() + "/js-reststubs" + suffix + ".zip");
        zipper.process();
    }

    private static void separateDocFiles(File coreStubs, File domStubs, File restStubs) {
        separateRestStubs(restStubs);
        separateCoreStubs(coreStubs);
        separateDomStubs(domStubs);
    }

    private static void separateCoreStubs(File coreStubsOutput) {
        File[] coreStubs = output.listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return name.contains("_CORE_");
            }
        });
        for (File file : coreStubs) {
            file.renameTo(new File(coreStubsOutput, file.getName()));
        }
    }

    private static void separateDomStubs(File domStubsOutput) {
        File[] domStubs = output.listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                return name.contains("_DOM_");
            }
        });
        for (File file : domStubs) {
            file.renameTo(new File(domStubsOutput, file.getName()));
        }
    }

    private static void separateRestStubs(File restStubsOutput) {
        File[] restStubs = output.listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                List<String> restStubs = new LinkedList<String>();
                restStubs.add("stub_CORE_Arguments.js");
                restStubs.add("stub_CORE_XMLHttpRequest.js");
                restStubs.add("stub_DOM_Console.js");
                restStubs.add("stub_DOM_Crypto.js");
                restStubs.add("stub_DOM_document.js");
                restStubs.add("stub_DOM_History.js");
                restStubs.add("stub_DOM_Location.js");
                restStubs.add("stub_DOM_Navigator.js");
                restStubs.add("stub_DOM_Screen.js");
                restStubs.add("stub_DOM_Storage.js");
                restStubs.add("stub_DOM_Window.js");
                restStubs.add("stub_DOM_window.js");
                return restStubs.contains(name);
            }
        });
        for (File file : restStubs) {
            file.renameTo(new File(restStubsOutput, file.getName()));
        }
    }
}
