package org.netbeans.jsstubsgenerator.core.ecma;

import java.io.File;
import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;
import org.netbeans.jsstubsgenerator.core.mdn.ReturnType;
import org.netbeans.jsstubsgenerator.stub.*;

/**
 *
 * @author Martin Fousek <marfous@netbeans.org>
 */
public class EcmaCoreGenerator {

    private static final Logger LOGGER = Logger.getLogger(EcmaCoreGenerator.class.getName());

    public static Map<String, StubObject> getUpdatedStubs(File inputDir, Map<String, StubObject> allStubs) throws IOException {
        File inputFile = new File(inputDir, "core.html");
        if (!inputFile.exists()) {
            LOGGER.log(Level.WARNING, "Dir with ecma core spec {0} doesn''t exist!", inputFile.getPath());
        }

        return getStubsAfterProcessedEcmaSpec(inputFile, allStubs);
    }

    private static Map<String, StubObject> getStubsAfterProcessedEcmaSpec(File inputFile, Map<String, StubObject> allStubs) throws IOException {
        // parse the specification
        Document doc = Jsoup.parse(inputFile, "UTF-8");

        // new map of core objects created on base of parsed ecma specification
        Map<String, StubObject> ecmaStubs = getEcmaGlobalEntries(doc, allStubs);

        // go thru all global objects and prune its properties and methods
        for (Entry<String, StubObject> stubEntry : ecmaStubs.entrySet()) {
            StubObject globalStub = stubEntry.getValue();
            globalStub.setPrune(true);
            // non-object stubs are already checked above
            if (globalStub.getType() != Type.OBJECT) {
                continue;
            }
            filterEcmaObjectEntries(doc, globalStub);
        }

        // complete missing attributes like escape, unescape - issue #215917
        for (Element element : getAllEcmaObjects(doc, new String[]{"B.2.", "15.1."})) {
            String elementName = element.ownText().indexOf(" ") > 0 ? element.ownText().substring(0, element.ownText().indexOf(" ")) : element.ownText();
            if (!ecmaStubs.containsKey(elementName)) {
                if (isEcmaMethod(element.ownText())) {
                    if (isGlobalMethod(element.ownText())) {
                        StubObject so = new StubObject(Origin.CORE, elementName, StubObject.GLOBAL);
                        so.setType(Type.METHOD);
                        so.setSyntax(element.ownText().replace(" ", ""));
                        tryToAttachParameters(so, element.ownText());
                        so.setReturns(new Return(ReturnType.guessReturns("global", elementName, ""), ""));
                        ecmaStubs.put(elementName, so);
                    } else {
                        //NOT NECESSARY AT ECMA 5.1
//                        StubObject so = new StubObject(Origin.CORE, elementName, getMethodParent(element.ownText()));
                    }
                }
            }
        }

        return ecmaStubs;
    }

    private static void filterEcmaObjectEntries(Document doc, StubObject globalStub) {
        Map<String, StubObject> ecmaAttributes = new HashMap<String, StubObject>();
        String objectDefiningSection = getObjectDefiningSection(doc, globalStub.getName());
        assert objectDefiningSection != null; // all sections has to be found in the specification

        // search in the parent section
        objectDefiningSection = objectDefiningSection.split("[ ]")[0];
        String searchSection = "";
        String[] sections = objectDefiningSection.split("[.]");
        for (int i = 0; i < sections.length - 1; i++) {
            searchSection += sections[i] + ".";
        }

        for (Entry<String, StubObject> attrEntry : globalStub.getAttributes().entrySet()) {
            StubObject attrStub = attrEntry.getValue();
            attrStub.setPrune(true);
            if (!validAndUpdate(doc, searchSection, globalStub.getName(), attrStub)
                    && !isValidatedAdditionally(globalStub.getName(), attrStub)) {
                LOGGER.log(Level.INFO, "Removed entry: {0}.{1}", new Object[]{attrStub.getParent(), attrStub.getName()});
//                System.err.println("-------- " + attrStub.getParent() + ":" + attrStub.getName());
            } else {
                ecmaAttributes.put(attrStub.getName(), attrStub);
            }
        }
        globalStub.setAttributes(ecmaAttributes);
    }

    private static boolean validAndUpdate(Document doc, String sectionNumber, String parent, StubObject attr) {
        // classes which inherit from Error are OK
        if (parent.endsWith("Error") && !"Error".equals(parent)) {
            return true;
        }

        Elements sectionElementHeaders = getAllEcmaObjects(doc, new String[]{sectionNumber, "B.2."});
        for (Element element : sectionElementHeaders) {
            String headerText = element.ownText();
            if (headerText.contains(attr.getName() + "(")
                    || headerText.contains(attr.getName() + " ")
                    || headerText.endsWith(attr.getName())) {
                updateSyntax(attr, element);
                return true;
            }
        }
        return false;
    }

    private static Elements getAllEcmaObjects(Document doc, String[] sections) {
        Elements result = new Elements();
        for (String section : sections) {
            result.addAll(getEcmaObjectOfSection(doc, section));
        }
        return result;
    }

    private static Elements getEcmaObjectOfSection(Document doc, String section) {
        return doc.select("section[id^=sec-" + section + "] > h1");
    }

    private static void updateSyntax(StubObject attr, Element element) {
        String headerString = element.text();
        String codeName = headerString.substring(headerString.indexOf(attr.getName()));
        String codeNameNoWS = codeName.replace(" ", "");
        attr.setSyntax(codeNameNoWS);
    }

    private static Map<String, StubObject> getEcmaGlobalEntries(Document doc, Map<String, StubObject> allStubs) {
        Map<String, StubObject> stubs = new HashMap<String, StubObject>();
        Iterator<Entry<String, StubObject>> iterator = allStubs.entrySet().iterator();
        while (iterator.hasNext()) {
            Entry<String, StubObject> stub = iterator.next();
            StubObject stubObject = stub.getValue();

            // work only with entries accessible in ecma spec - remove global entries
            if (isDefinedInGlobalScope(doc, stubObject.getName())) {
                // set to the stub that it must be pruned
                stubObject.setPrune(true);
                stubs.put(stubObject.getName(), stubObject);
            } else {
                LOGGER.log(Level.INFO, "Removed global object: {0}", stubObject.getName());
            }
        }
        return stubs;
    }

    private static String getObjectDefiningSection(Document doc, String objectName) {
        Elements objectsSections = doc.select("section[id^=sec-15.1.4.] > h1");
        objectsSections.addAll(doc.select("section[id^=sec-15.1.5.] > h1"));

        for (Element header : objectsSections) {
            // find in chapter 15.1.4.x, 15.1.5.x the section numbers
            if (header.ownText().contains(objectName + " ") || header.ownText().endsWith(objectName)) {
                String lastSection = "";
                for (Element section : header.nextElementSibling().select("a")) {
                    lastSection = section.text();
                }
                return lastSection;
            }
        }
        return null;
    }

    private static boolean isDefinedInGlobalScope(Document doc, String name) {
        Elements section15ElementHeaders = doc.select("section[id^=sec-15.1.] > h1");
        for (Element header : section15ElementHeaders) {
            String headerText = header.ownText();
            // chapter headers
            if (headerText.split("[ ]").length >= 2 && !headerText.contains("(")) {
                continue;
            } else {
                // check for name (ignore brackets, etc.)
                if (headerText.contains(name)) {
                    return true;
                }
            }
        }
        return false;
    }

    private static boolean isEcmaMethod(String ownText) {
        return ownText.contains("(");
    }

    private static boolean isGlobalMethod(String ownText) {
        return !ownText.contains(".prototype.");
    }

    private static String getMethodParent(String ownText) {
        return ownText.substring(0, ownText.indexOf(".prototype"));
    }

    private static void tryToAttachParameters(StubObject so, String ownText) {
        int indexLeft = ownText.indexOf("(");
        int indexRight = ownText.indexOf(")");
        if (indexLeft == -1 || indexRight == -1) {
            return;
        }
        
        String params = ownText.substring(indexLeft + 1, indexRight);
        if (params.trim().isEmpty()) {
            return;
        }

        for (String param : params.split("[,]+")) {
            if (!param.trim().isEmpty()) {
                so.addParameter(new Parameter(param.trim(), ReturnType.guessReturns("global", param.trim(), ""), ""));
            }
        }
    }

    private static boolean isValidatedAdditionally(String parent, StubObject attribute) {
        // see issue #215918
        Map<String,Set<String>> allowedEntries = new HashMap<String,Set<String>>();
        Set<String> stringItems = new HashSet<String>(Arrays.asList("anchor", "big", "blink", "bold", "fixed",
                "fontcolor", "fontsize", "italics", "link", "small", "strike", "sub", "sup"));
        allowedEntries.put("String", stringItems);

        Set set = allowedEntries.get(parent);
        if (set != null) {
            return set.contains(attribute.getName());
        } else {
            return false;
        }
    }
}
