package org.netbeans.jsstubsgenerator.core.mdn;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.jsoup.Jsoup;
import org.jsoup.nodes.Document;
import org.jsoup.nodes.Element;
import org.jsoup.select.Elements;
import org.netbeans.jsstubsgenerator.stub.*;

/**
 *
 * @author Martin Fousek <marfous@netbeans.org>
 */
public class MdnGenerator {

    private static final Logger LOGGER = Logger.getLogger(MdnGenerator.class.getName());
    private static final FilenameFilter CORE_FILE_FILTER = new CoreJsFileFilter();
    private static final FilenameFilter DOM_FILE_FILTER = new DomJsFileFilter();
    private static final String JS_CORE_DIR = "JavaScript/Reference/Global_Objects";
    private static final String JS_DOM_DIR = "DOM";

    public static Map<String, StubObject> parseCoreDocs(File inputDir) throws IOException {
        Map<String, StubObject> stubs = new HashMap<String, StubObject>();
        File dirWithCoreStubs = new File(inputDir, JS_CORE_DIR);
        if (!dirWithCoreStubs.exists()) {
            LOGGER.log(Level.WARNING, "Dir with core docs {0} doesn''t exist!", dirWithCoreStubs.getPath());
        }

        loadCoreStubs(dirWithCoreStubs, stubs);
        enhanceReturnTypes(stubs);
        return stubs;
    }

    public static Map<String, StubObject> parseDomDocs(File inputDir) throws IOException {
        Map<String, StubObject> stubs = new HashMap<String, StubObject>();
        File dirWithDomStubs = new File(inputDir, JS_DOM_DIR);
        if (!dirWithDomStubs.exists()) {
            LOGGER.log(Level.WARNING, "Dir with DOM docs {0} doesn''t exist!", dirWithDomStubs.getPath());
        }

        loadDomStubs(dirWithDomStubs, stubs);
        enhanceReturnTypes(stubs);
        return stubs;
    }

    private static void attachSummary(StubObject stubObject, File summaryFile, boolean object) throws IOException {
        Document summaryDoc = Jsoup.parse(summaryFile, "UTF-8");
        if (summaryDoc.select("h1").text().contains("__")) {
            stubObject.setUnderscored(true);
        }
        Elements divs = summaryDoc.select("div[class=pageText] > div");

        if (object) {
            stubObject.setType(Type.OBJECT);
        } else {
            String tableContent = divs.select("table[class=standard-table]").text();
            if (tableContent.contains("Property")) {
                stubObject.setType(Type.PROPERTY);
            } else {
                stubObject.setType(Type.METHOD);
            }
        }

        if (divs.isEmpty()) {
            // when the page was redirected to home, no doc exists for this attribute - it should be object
            if (!"uneval".equals(stubObject.getName())) {
                stubObject.setType(Type.OBJECT);
            }
            stubObject.setUncompleted(true);
            return;
        }

        // summary, syntax, description, compatibility, examples, properties, ...
        attachSummaryForHeaders(stubObject, divs, "h2", object);
        // due to pages like Error, we have to try both header elements
//        if ("".equals(stubObject.getSummary())) {
        attachSummaryForHeaders(stubObject, divs, "h3", object);
//        }

    }

    private static void attachSummaryForHeaders(StubObject stubObject, Elements divs, String headerType, boolean object) {
        for (Element div : divs) {
            if (div.select(headerType).text().equals("Summary")) {
                // summary
                stubObject.setSummary(div.select("p").text());
                continue;
            } else if (div.select(headerType).text().equals("Syntax")) {
                // syntax
                String syntax = div.select("p").html();
                if ("".equals(syntax)) {
                    // i.e. encodeURIComponent - try PRE element
                    syntax = div.select("pre").html();
                }
                if ("".equals(syntax)) {
                    // i.e. encodeURIComponent - try PRE element
                    syntax = div.select("code").html();
                }
                syntax = syntax.replace("<br>", "NNNNN");
                syntax = syntax.replace("<br/>", "NNNNN");
                syntax = syntax.replace("<br >", "NNNNN");
                syntax = syntax.replace("<br />", "NNNNN");
                syntax = Jsoup.parse(syntax).text().trim();
                syntax = syntax.replace("NNNNN", "<br>");
                stubObject.setSyntax(syntax);
                continue;
            } else if (div.select(headerType).text().equals("Description")) {
                // description
                stubObject.setDescription(div.select("p").html());
                continue;
            } else if (div.select(headerType).text().equals("Parameters") || div.select(headerType).text().equals("Parameter")) {
                // parameters
                for (Element dl : div.select("dl")) {
                    for (Element dt : dl.select("dt")) {
                        String name = dt.text().trim().replace(" ", "");
                        String description = dt.nextElementSibling() == null ? "" : dt.nextElementSibling().text();
                        if (stubObject.getName().equals("Reduce")) {
                            if (!"callback".equals(name) && !"initialValue".equals(name)) {
                                continue;
                            }
                        } else if (stubObject.getName().equals("test") || stubObject.getName().equals("exec")) {
                            if ("regexp".equals(name)) {
                                continue;
                            }
                        } else if (stubObject.getName().equals("pow")) {
                            if ("exponents".equals(name)) {
                                name = "exponent";
                            }
                        } else if (stubObject.getName().equals("atan2")) {
                            if ("y, x".equals(name)) {
                                stubObject.addParameter(new Parameter(
                                        "y",
                                        ReturnType.guessReturns(stubObject.getName(), name, description),
                                        description));
                                stubObject.addParameter(new Parameter(
                                        "x",
                                        ReturnType.guessReturns(stubObject.getName(), name, description),
                                        description));
                                continue;
                            }
                        }
                        stubObject.addParameter(new Parameter(
                                name,
                                ReturnType.guessReturns(stubObject.getName(), name, description),
                                description));
                    }
                }
                continue;
            }
            if (object) {
                if (div.select(headerType).text().equals("Properties")) {
                    // class properties
                    List<String> properties = getAttributes(div);
                    if (properties.isEmpty()) {
                        LOGGER.log(Level.WARNING, "Empty list of properties for {0}", stubObject);
                    }
                    stubObject.putProperties(properties);
                    continue;
                } else if (div.select(headerType).text().equals("Methods")) {
                    // class methods
                    List<String> methods = getAttributes(div);
                    if (methods.isEmpty()) {
                        LOGGER.log(Level.WARNING, "Empty list of methods for {0}", stubObject);
                    }
                    stubObject.putMethods(methods);
                    continue;
                } else if (div.select(headerType).text().trim().endsWith("instances")) {
                    for (Element innerDiv : div.select("div")) {
                        attachInstanceAttributesForHeaders(stubObject, innerDiv, "h3");
                        attachInstanceAttributesForHeaders(stubObject, innerDiv, "h4");
                    }
                    continue;
                } else if (!div.select(headerType).isEmpty()
                        && div.select(headerType).get(0).text().trim().endsWith("instances")) {
                    // in cases of headers for instances/methods/props are in the same div - Error page
                    attachInstanceAttributesForHeaders(stubObject, div, "h3");
                }
            }
        }
    }

    private static List<String> getAttributes(Element div) {
        List<String> attributes = new LinkedList<String>();
        Elements elements = div.select("dl > dt > a");
        elements.addAll(div.select("p > a"));
        elements.addAll(div.select("dl > dt > code > a"));
        // DOM Element
        elements.addAll(div.select("tr > td > code > code > a"));

        for (Element element : elements) {
            String attr = element.text();
            if (!attr.contains(" ") && !attr.contains("?")) {
                // more words - it doesn't appear to be valid attribute
                attributes.add(attr);
            }
        }

        return attributes;
    }

    private static List<String> getAttributes(Elements wrappers) {
        List<String> attributes = new LinkedList<String>();
        for (Element wrapper : wrappers) {
            attributes.addAll(getAttributes(wrapper));
        }

        return attributes;
    }

    private static void attachInstanceAttributesForHeaders(StubObject stubObject, Element innerDiv, String tagType) {
        boolean found = false;
        if (innerDiv.select(tagType).text().equals("Properties")) {
            // instance properties
            List<String> properties = getAttributes(innerDiv);
            if (properties.isEmpty()) {
                LOGGER.log(Level.WARNING, "Empty list of instance properties for {0}", stubObject);
            }
            found = true;
            stubObject.putInstanceProperties(properties);
        } else if (innerDiv.select(tagType).text().equals("Methods")) {
            // instance methods
            List<String> methods = getAttributes(innerDiv);
            if (methods.isEmpty()) {
                LOGGER.log(Level.WARNING, "Empty list of instance methods for {0}", stubObject);
            }
            found = true;
            stubObject.putInstanceMethods(methods);
        }
        if (!found) {
            // i.e. Error case
            Elements dls = innerDiv.select("h3:contains(Properties) ~ dl");
            if (!dls.isEmpty()) {
                List<String> properties = getAttributes(dls);
                if (properties.isEmpty()) {
                    LOGGER.log(Level.WARNING, "Empty list of instance properties for {0}", stubObject);
                }
                stubObject.putInstanceProperties(properties);
            }
            dls = innerDiv.select("h3:contains(Methods) ~ dl");
            if (!dls.isEmpty()) {
                List<String> methods = getAttributes(dls);
                if (methods.isEmpty()) {
                    LOGGER.log(Level.WARNING, "Empty list of instance properties for {0}", stubObject);
                }
                stubObject.putInstanceMethods(methods);
            }
        }
    }

    private static void loadCoreStubs(File dirWithCoreStubs, Map<String, StubObject> allStubs) throws IOException {
        for (File docFile : dirWithCoreStubs.listFiles(CORE_FILE_FILTER)) {
            // get stub name from the file name
            String className = docFile.getName();
            StubObject stubObject = new StubObject(Origin.CORE, className, StubObject.GLOBAL);

            if (allStubs.get(className) != null) {
                LOGGER.log(Level.WARNING, "Stub entry already exists: ", className);
            }

            // objects are defined in directories
            if (docFile.isDirectory()) {
                // check existency of the summary file
                File summaryFile = new File(docFile, docFile.getName());
                if (!summaryFile.exists()) {
                    LOGGER.log(Level.WARNING, "Summary for stup does not exist: ", summaryFile.getPath());
                } else {
                    attachSummary(stubObject, summaryFile, true);
                }

                // parse all properties, methods, ...
                for (File attrFile : docFile.listFiles(CORE_FILE_FILTER)) {
                    String attrName = attrFile.getName();
                    StubObject attrObject = new StubObject(Origin.CORE, attrName, className);
                    attachSummary(attrObject, attrFile, false);

                    stubObject.assignAttribute(attrName, attrObject);
                }
                // assign appropriate type
                for (Map.Entry<String, StubObject> entry : stubObject.getAttributes().entrySet()) {
                    StubObject attribute = entry.getValue();
                    attribute.setType(attribute.getType());
                }
            } else {
                attachSummary(stubObject, docFile, false);
            }

            allStubs.put(className, stubObject);
        }
    }

    private static String getParentFromFileName(String filename) {
        int dotOffset = filename.indexOf(".");
        String parentName = filename.substring(0, dotOffset);
        if (!parentName.isEmpty()) {
            return parentName.substring(0, 1).toUpperCase() + parentName.substring(1);
        }
        return parentName;
    }

    private static String getFieldName(String filename) {
        int dotOffset = filename.indexOf(".");
        String fieldName = filename.substring(dotOffset + 1, filename.length());
        if (fieldName.contains("?")) {
            return null;
        }
        return fieldName;
    }

    private static void loadDomStubs(File dirWithDomStubs, Map<String, StubObject> allStubs) throws IOException {
        for (File file : dirWithDomStubs.listFiles(DOM_FILE_FILTER)) {
            String fileName = file.getName();
            String parentElementName = getParentFromFileName(fileName);
            StubObject parentStub = allStubs.get(parentElementName);
            File summaryFile = new File(file.getParentFile(), parentElementName);
            if (parentStub == null) {
                parentStub = new StubObject(Origin.DOM, parentElementName, StubObject.GLOBAL);
                parentStub.setPrune(true);
                if (summaryFile.exists()) {
                    attachSummary(parentStub, summaryFile, true);
                }
            }

            // get name from the file name
            String fieldName = getFieldName(fileName);
            if (fieldName == null) {
                continue;
            }
            StubObject attribute = parentStub.getAttribute(fieldName);
            if (attribute != null) {
                attribute.setOrigin(Origin.DOM);
                attribute.setPrune(true);

                // try to attach return type
                if (summaryFile.exists()) {
                    Document summaryDoc = Jsoup.parse(summaryFile, "UTF-8");
                    Elements elements = summaryDoc.select("tr:contains(" + fieldName + ")");
                    if (!elements.isEmpty()) {
                        if (attribute.getType() == Type.PROPERTY) {
                            String returnType = trimReturnType(elements.select("td:eq(2)").text());
                            String returnDesc = elements.select("td:eq(1)").text();
                            attribute.setReturns(new Return(returnType, returnDesc));
                        } else if (attribute.getType() == Type.METHOD) {
                            String returnType = trimReturnType(elements.select("td:eq(1)").text());
                            String returnDesc = elements.select("td:eq(0)").text();
                            attribute.setReturns(new Return(returnType, returnDesc));
                        }
                    }
                }
            }

            allStubs.put(parentElementName, parentStub);
        }
    }

    private static String trimReturnType(String returnString) {
        int offset = returnString.indexOf(" ");
        String returnType = offset == -1 ? returnString : returnString.substring(0, offset);
        if ("-".equals(returnType)) {
            return "undefined";
        }
        return returnType;
    }

    private static void enhanceReturnTypes(Map<String, StubObject> allStubs) {
        for (Map.Entry<String, StubObject> entry : allStubs.entrySet()) {
            StubObject stubObject = entry.getValue();
            enhanceReturnType(stubObject);
            for (Map.Entry<String, StubObject> attrEntry : stubObject.getAttributes().entrySet()) {
                enhanceReturnType(attrEntry.getValue());
            }
        }
    }

    private static void enhanceReturnType(StubObject stubObject) {
        String returns = ReturnType.serachFor(stubObject.getParent(), stubObject.getName());
        if (returns != null) {
            if (stubObject.getType() == Type.OBJECT && !stubObject.getName().equals(returns)) {
                stubObject.setSuperClass(returns);
            }
            stubObject.setReturns(new Return(returns, null));
        }
    }
}
