/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 2010 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 *
 * Contributor(s):
 *
 * Portions Copyrighted 2010 Sun Microsystems, Inc.
 */

package org.netbeans.server.uihandler.api.bugs;

import java.io.IOException;
import java.util.Collection;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.persistence.EntityManager;
import javax.servlet.AsyncContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import org.netbeans.modules.exceptions.entity.Report;
import org.netbeans.modules.exceptions.entity.Submit;
import org.netbeans.modules.exceptions.utils.PersistenceUtils;
import org.netbeans.modules.exceptions.utils.ShortTimeCache;
import org.netbeans.server.uihandler.api.Issue;
import org.netbeans.server.utils.Persistable;
import org.netbeans.server.utils.Persistable.TransactionResult;

/**
 *
 * @author Jindrich Sedek
 */
public class BugReporterFacade {

    public static final Logger LOG = Logger.getLogger("org.netbeans.modules.exceptions.bugs.reporter");// NOI18N

    private final BugReporter delegate;
    private static final ShortTimeCache<Integer, Issue> issueCache = new ShortTimeCache<Integer, Issue>();
    
    public BugReporterFacade (BugReporter delegate) {
        this.delegate = delegate;
    }

    public void reportSubmit(Submit submit, String password) {
        reportSubmit(submit, password, null, null);
    }

    public void reportSubmit(Submit submit, String password, String extraComment) {
        reportSubmit(submit, password, extraComment, null);
    }

    public void reportSubmit(final Submit submit, final String password, final String extraComment, final String extraCC) {
        reportSubmit(submit, password, null, extraComment, extraCC, false);
    }
    
    // this class is Singleton thanks to BugReporterFactory and thus a report will be reported only once
    public synchronized void reportSubmit(final Submit submit, final String password, 
            final AsyncContext asyncContext,
            final String extraComment, final String extraCC, final boolean isReopen) {
        final int id = submit.getId();
        
        PersistenceUtils.processPersistable(new Persistable.Transaction() {

            public TransactionResult runQuery(EntityManager em) {
                try {
                    Submit thisSubmit = em.find(submit.getClass(), id);
                    if (thisSubmit == null) {
                        throw new IllegalStateException("Submit id " + id + " was not found"); // NOI18N
                    }
                    Report report = thisSubmit.getReportId();
                    if (!isReopen && report.isInIssuezilla()){
                        // stop insertion
                        return TransactionResult.ROLLBACK;
                    }
                    int bugId = delegate.reportSubmit(thisSubmit.getSubmit(em), password, extraComment, extraCC);
                    report.setIssueId(bugId, BugReporterFacade.class);
                    report.setIssuezillaTransfer(false);
                    report = em.merge(report);
                    if (asyncContext != null) {
                        processAsyncContextReply(asyncContext, thisSubmit.getId(), report.getId(), bugId, "INSERT");
                    }                    
                    return TransactionResult.COMMIT;
                } catch (Exception e) {
                    String message = "Bug was not created for report #" + id; // NOI18N
                    LOG.log(Level.SEVERE, message, e);
                    throw new BugReporterException(message, e);
                }
            }
            
            private void processAsyncContextReply(AsyncContext asyncContext, int submitId, int reportId, int bugNo, String action) {

                try {
                    LOG.log(Level.INFO, "Sending async servlet reply: {0}. Submit {1}.", new Object[] { action, submitId } ); // NOI18N 

                    ServletRequest request = asyncContext.getRequest();
                    ServletResponse response = asyncContext.getResponse();
                    request.setAttribute("action", action);
                    request.setAttribute("reportNo", reportId);
                    request.setAttribute("bugNo", bugNo);
                    request.getRequestDispatcher("/uploadredirect.jsp").forward(request, response); // NOI18N
                    asyncContext.complete();

                } catch (ServletException | IOException | IllegalStateException ex) {
                    // Illegal state : in case response has timed out - we still need to have a bug stored in db
                    //     although we have entered it late (after async server response http timeout in this case)
                    //     - thereofore no rethrow

                    // This is a not very clean code but there's really no check like "isComplete()" on AsyncContext API
                    // (Note: if there's one in the future, it needs to be synchronided)
                    if (!(ex instanceof IllegalStateException) || !ex.getMessage().contains("already completed")) {
                        LOG.log(Level.SEVERE, ex.getMessage(), ex);
                    }
                }
            }

        });
    }

    public void createAttachment(Submit submit, String password) throws BugReporterException{
        delegate.createAttachment(submit.getSubmit(null), password);
    }

    public void postTextComment(int issueId, String text, String username, String password) throws BugReporterException{
        delegate.postTextComment(issueId, text, username, password);
    }

    public void setIssuePriority(int issueId, String priority, String username, String password) throws BugReporterException{
        delegate.setIssuePriority(issueId, priority, username, password);
    }

    public void addToCCList(int issueId, String username, String password) throws BugReporterException{
        delegate.addToCCList(issueId, username, password);
    }

    public void reopenIssue(int issueId, String comment, String username, String password) throws BugReporterException{
        delegate.reopenIssue(issueId, comment, username, password);
    }

    public Issue getIssue(int issueId){
        Issue is = issueCache.get(issueId);
        if (is == null){
            is = delegate.getIssue(issueId);
            issueCache.put(issueId, is);
        }
        return is;
    }

    public Collection<Integer> filterOpen(Collection<Integer> ids){
        return delegate.filterOpen(ids);
    }

    public void postTextComment(int suezillaId, String commentMessage) {
        postTextComment(suezillaId, commentMessage, null, null);
    }
    
    public String getManyDuplicatesComment(EntityManager em, Submit exc) {
        return delegate.getManyDuplicatesComment(em, exc);
    }

    public void reopenIssue(int suezillaId, String commentMessage) {
        reopenIssue(suezillaId, commentMessage, null, null);
    }

    public void setIssuePriority(int suezillaId, String priorityStr) {
        setIssuePriority(suezillaId, priorityStr, null, null);
    }
    
    public void addOrUpdateTrackingGroup(int suezillaId, String trackingGroupName, String trackingGroupValue) {
        delegate.addOrUpdateTrackingGroup(suezillaId, trackingGroupName, trackingGroupValue);
    }

    public void clearCache(){
        issueCache.clear();
    }

    /*** Following methods provide interface to support admin web console ***/
    public void adminEnsureConnectionAlive() {
        delegate.ensureConnectionAlive();
    }

    public int adminGetNewBugStatus() {
        return delegate.getNewBugStatus();
    }

    public void adminSetNewBugStatus(int newBugStatus) {
        delegate.setNewBugStatus(newBugStatus);
    }
}
