/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.server.uihandler;

import java.io.File;
import java.io.IOException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import org.openide.util.NbBundle;
import org.netbeans.server.uihandler.api.Issue;
import org.netbeans.server.snapshots.SlownessChecker;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;
import java.util.regex.Pattern;
import javax.persistence.EntityManager;
import javax.persistence.Query;
import javax.servlet.AsyncContext;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import org.netbeans.modules.exceptions.utils.PersistenceUtils;
import org.netbeans.modules.exceptions.utils.Component;
import org.netbeans.server.componentsmatch.Matcher;
import org.netbeans.modules.exceptions.entity.*;
import org.netbeans.modules.exceptions.entity.BuildInfo.Vcs;
import org.netbeans.modules.exceptions.entity.Slowness.SlownessType;
import org.netbeans.modules.exceptions.utils.LogfileTools;
import org.netbeans.server.uihandler.api.bugs.BugReporterFacade;
import org.netbeans.server.uihandler.api.bugs.BugReporterFactory;
import org.netbeans.server.services.beans.Components;
import static org.netbeans.server.uihandler.Utils.*;
import org.netbeans.server.uihandler.api.bugs.BugTrackerCategorizer;
import org.netbeans.server.uihandler.api.bugs.BugTrackerProvidersFactory;
import org.netbeans.server.uihandler.api.bugs.BugTrackerReopenChecker;
import org.netbeans.server.uihandler.api.bugs.BuildQuery;
import org.netbeans.server.uihandler.api.bugs.ReopenCallback;
import org.netbeans.server.uihandler.api.bugs.ReportThrottle;
import org.netbeans.server.uihandler.api.bugs.ReportThrottleFactoryProvider;
import org.openide.util.RequestProcessor;

/**
 *
 * @author Jindrich Sedek
 */
final class DbInsertion {
    
    private static final Logger LOG = Logger.getLogger("org.netbeans.modules.exceptions.entity.Report.IssueChanges"); // NOI18N
    private static Logger INSERTION_LOG = Logger.getLogger("org.netbeans.server.uihandler.DbInsertion"); // NOI18N
    private static Logger MATCHER_LOG = Logger.getLogger("org.netbeans.server.uihandler.matcher"); // NOI18N
    private static final int SERVER_BUSY_REPLY_IN_SECONDS = 45;    
    private static final int MAXMESSAGELENGTH = 1024;
    private static final int CHANGESET_RADIX = 16;
    private static final Component DEFAULT_JDEV_COMPONENT;
    private static final Component DEFAULT_DCS_COMPONENT;
    private static final int DCS_PRODUCT = 10427;
    static {
        DEFAULT_JDEV_COMPONENT = new Component("JDEV", ""); // NOI18N
        DEFAULT_JDEV_COMPONENT.setProduct(807);

        DEFAULT_DCS_COMPONENT = new Component("TESTS", ""); // NOI18N
        DEFAULT_DCS_COMPONENT.setProduct(DCS_PRODUCT);
    }
    private static final int CHANGESET_IDX = 5;
    private static final String UNKNOWN = "unknown";// NOI18N
    private static final Pattern LABEL_PATTERN = Pattern.compile("([a-zA-Z]+)_([a-zA-Z]+)_([a-zA-Z]+)_([0-9]+)\\.([0-9]+)\\.([0-9]+)\\.?(S)"); // NOI18N
    private static final SimpleDateFormat labelDateFormat = new SimpleDateFormat("yyMMdd"); // NOI18N
    
    private final LogRecord userMetaData;
    private final PreparedParams preparedParams;    
    private final LogRecord buildInfo;
    private final Throwable thrown;
    private final LogFileTask thisTask;
    private final Long slownessTime;
    private final SlownessType slownessType;
    private final String latestAction;
    private int recursionDepth;
    private Integer submitId;
    private Report duplicatesReport;
    private BuildInfo naBuildInfo;
    

    private static final int CLEANUP_INITIAL_DELAY = 1000;
    private static final int MAX_CLEANUP_ATTEMPTS = 3;
    
    private static final int CLEANUP_TASK_INTERVAL = 100;
    private final BugReporterFacade bugReporter;
    private final BugTrackerCategorizer categorizer;        

    private static final ScheduledExecutorService scheduledExecutor;
    
    static {
        scheduledExecutor = Executors.newSingleThreadScheduledExecutor();
    }
    
    DbInsertion(LogRecord data, Throwable thrown, LogRecord buildInfo, Long slownessTime, SlownessType slownessType, String latestAction) {
        this(data, thrown, buildInfo, null, slownessTime, slownessType, latestAction);
    }

    DbInsertion(LogRecord data, Throwable thrown, LogFileTask task) {
        this(data, thrown, null, task, null, null, null);
    }
    
    DbInsertion(LogRecord data, Throwable thrown, LogRecord buildInfo, LogFileTask task) {
        this(data, thrown, buildInfo, task, null, null, null);
    }
    
    DbInsertion(LogRecord data, Throwable thrown, LogRecord buildInfo, LogFileTask task, Long slownessTime, SlownessType slownessType, String latestAction) {
        this.thrown = thrown;
        this.userMetaData = data;
        this.preparedParams = PreparedParams.prepare(userMetaData.getParameters(), thrown);
        this.buildInfo = buildInfo;
        this.slownessTime = slownessTime;
        this.latestAction = latestAction;
        this.slownessType = slownessType;
        if (task == null) {
            this.thisTask = LogFileTask.getRunningTask();
        } else {
            this.thisTask = task;
        }
        categorizer = BugTrackerProvidersFactory.getInstance().getCategorizer(preparedParams.getReportType());
        bugReporter = BugReporterFactory.getReporter();

    }
    
    private static final Timer timer = new Timer("Throttled Log cleanup"); // NOI18N
    private static final Queue<DbInsertion.CleanupRequest> cleanupQueue = new LinkedList<DbInsertion.CleanupRequest>(); 
    private static final RequestProcessor cleanupRequestProcessor = new RequestProcessor("Throttled Log cleanup", 1); // NOI18N
    private static final AtomicBoolean isCleanupProcessing = new AtomicBoolean(false);

    private static class CleanupRequest implements Runnable {
        private final Logfile logInfo;
        private final Report report;
        private final int attempt;
        
        private CleanupRequest(Logfile logInfo, Report report) {
            this(logInfo, report, 0);
        }
        
        private CleanupRequest(Logfile logInfo, Report report, int attempt) {
            this.logInfo = logInfo;
            this.report = report;
            this.attempt = attempt;
        }

        private void deleteIfExists(File f) {
            if (f.exists()) {
                f.delete();
                INSERTION_LOG.log(Level.INFO, "Cleanup: Deleted {0}.", f.getPath());
            } else {
                INSERTION_LOG.log(Level.INFO, "Cleanup: Couldn't delete {0}, file was not found.", f.getPath());
            }
        }

        private void doCleanup() {
            INSERTION_LOG.log(Level.INFO, "Starting throttle cleanup task for {0}, {1}", // NOI18N
                new Object[] { report.getId(), logInfo.getFileName() });
            deleteIfExists(LogfileTools.getMessagesFile(logInfo));
            deleteIfExists(LogfileTools.getNPSFile(logInfo));
            deleteIfExists(LogfileTools.getHeapDumpFile(logInfo));
            deleteIfExists(LogfileTools.getEventsFile(logInfo));
            deleteIfExists(LogfileTools.getAddInfoFile(logInfo));
            // delete this file as last to prevent race condition with uploader
            deleteIfExists(LogfileTools.getLoggerFile(logInfo));
            // Note: in the case the logger file is uploaded and others are not, there may be some files left after cleanup. 
            //       However this should be extremely rare and even if it happens, these files would be overwritten with the
            //       next upload from same reporter (as the main loggerFile doesn't exist anymore)
            
            if (report != null) {
                // currently sync is probably not needed, CleanupRequest run on a single queue
                    // (see if we can use some row locking via hibernate)
                synchronized (CleanupRequest.class) {
                    // Not sure if entity manager should be there just once for class, not create it again fo every request
                    //       - study guidelines of JPA (hibernate)
                    EntityManager em = PersistenceUtils.getInstance().createEntityManager();
                    em.getTransaction().begin();
                    Integer throttledCount = report.getThrottled();
                    int throttledInt = (throttledCount == null) ? 0 : throttledCount.intValue() + 1;
                    report.setThrottled(throttledInt);
                    em.merge(report);                    
                    INSERTION_LOG.log(Level.INFO, "Report {0} has now {1} throttled exceptions", // NOI18N
                        new Object[] { report.getId().toString(), throttledInt });
                    em.getTransaction().commit();
                }
            }
        }
        
        @Override
        public void run() {
            if (LogfileTools.getLoggerFile(logInfo).exists()) {
                // Synchronisation issue? Lof Info exists, we do delete, but other deletes won't perform (leaving orphans?)
                doCleanup();
            } else {
                // does not exist yet, try again later:
                if (attempt < MAX_CLEANUP_ATTEMPTS) {
                    cleanupRequestProcessor.post(new CleanupRequest(logInfo, report, attempt+1), attempt * CLEANUP_INITIAL_DELAY);
                }
            }
        }
    }
    
    private static class CleanupTask extends TimerTask {

        public void run() {
            DbInsertion.CleanupRequest firstRequest = cleanupQueue.poll();
            if (firstRequest != null) {
                cleanupRequestProcessor.post(firstRequest);
            }
        }
    }

    /** issueId is an ID of newly created issue if not duplicate
     *  if duplicate issueId is an ID of issue to be duplicate of
     *  issuezillaId is 0 if not assigned, otherwise an issuezilla Id
     * @return {int issueId, bool isDuplicate, int issuezillaId} or null if it's not an error report
     *
     */
    public void start(ExceptionsData data) {
        
        INSERTION_LOG.info("Starting DbInsertion task"); // NOI18N
        
        long time = System.currentTimeMillis(); 

        // Debug code
//        Long fixedBuildNumber = BuildQuery.findBuild(20930590);
//        System.out.println(fixedBuildNumber);
        // End of debug code
        
        ReportThrottle reportThrottle = ReportThrottleFactoryProvider.getReportThrottleFactory().getReportThrottle(preparedParams.getReportType());

        reportThrottle.resetParameters();
        reportThrottle.addParameter("testName", preparedParams.getTestName()); // NOI18N
        reportThrottle.addParameter("label", (String)buildInfo.getParameters()[1]); // NOI18N
        reportThrottle.addParameter("summary", preparedParams.getSummary()); // NOI18N
        
        if (reportThrottle.checkReportForThrottle(thisTask.getTaskEntityManager(), thrown)) {
            INSERTION_LOG.info("Throttling report"); // NOI18N
            processAsyncRequestReply(reportThrottle.getThrottledReport(), "THROTTLED");
            if(!isCleanupProcessing.getAndSet(true)){
                    timer.schedule(new DbInsertion.CleanupTask(), 0, CLEANUP_TASK_INTERVAL);
                }            
                cleanupQueue.add(new CleanupRequest(thisTask.getLogInfo(), reportThrottle.getThrottledReport()));
                
            INSERTION_LOG.log(Level.INFO, "Throttled issue for test name {0}, file {1}", // NOI18N
                new Object[] { preparedParams.getTestName(), thisTask.getLogInfo().getFileName() } );
                
            throw new InsertThrottledException();
        }
  
        // End of experimental throttling
        // Big question: can there be some nasty race condition because of new incomming log numbers?!?
        //      a) submitId ... PersistenceUtils.getNextId is thread save and is used only AFTER throttling
        //      b) uploadNumber -> this is generated inside servlet, it can 
        //                      possibly be not thread safe (should we synchronize RequestFacadeImpl.move()?)
        //                      but throttling won't affect this any further
        submitId = PersistenceUtils.getNextId(Submit.class);
        INSERTION_LOG.log(Level.INFO, "[sbm {0}] Processing new submit under submit id {0}", new Object[] { String.valueOf(submitId) } ); // NOI18N
        data.setSubmitId(submitId);
        insertLogFileInformation(submitId);
        Submit submit = null;
        if (slownessTime != null){
            data.setDataType(ExceptionsData.DataType.SLOWNESS_REPORT);
            SlownessChecker checker = new SlownessChecker(thisTask.getTaskEntityManager(), latestAction, slownessType, thisTask.getLogInfo());
            duplicatesReport = checker.checkSlowness();
            submit = insertSlownessData(checker);
        } else if (thrown != null) {
            data.setDataType(ExceptionsData.DataType.EXC_REPORT);
            long start = System.currentTimeMillis();
            String testName = preparedParams.getTestName();
            String reportType = preparedParams.getReportType();
            MATCHER_LOG.log(Level.INFO, "[sbm {0}] Trying to find duplicate for submit {0}", new Object[] { submitId.toString() }); // NOI18N
            if (testName != null && !(testName.isEmpty())) {
                duplicatesReport = categorizer.matchToReport(thisTask.getTaskEntityManager(), submitId, mergeThrowableHierarchicaly(thrown), testName, reportType);
            } else {
                duplicatesReport = categorizer.matchToReport(thisTask.getTaskEntityManager(), submitId, mergeThrowableHierarchicaly(thrown), preparedParams.getApiMark(), reportType);
            }
            if (duplicatesReport == null) {
                String msg = "[sbm {0}] No duplicate found for submit {0}"; // NOI18N
                MATCHER_LOG.log(Level.INFO, msg, new Object[] { submitId.toString() });
                INSERTION_LOG.log(Level.INFO, msg, new Object[] { submitId.toString() });
            } else {
                String msg = "[sbm {0}] Duplicate found for submit {0} : report {1} "; // NOI18N
                MATCHER_LOG.log(Level.INFO, msg, new Object[] { submitId, duplicatesReport.getId().toString() });
                INSERTION_LOG.log(Level.INFO, msg, new Object[] { submitId, duplicatesReport.getId().toString() });
            }
            long stop = System.currentTimeMillis() - start;
            if (stop > 2000 ){
                String id="No duplicate found"; // NOI18N
                if(duplicatesReport!=null){
                    id = duplicatesReport.getId().toString();
                }
                LOG.log(Level.WARNING, "[sbm " + submitId + "] Utils.checkIssue() took:" + stop + " for DR.id:" + id); // NOI18N
            }
            submit = insertException();
            data.setNewReportId(submit.getReportId().getId());
        } else {// only a log file insertion
            data.setDataType(ExceptionsData.DataType.STATISTIC);
        }
        
        if (submit != null){

            
        final int reportId = submit.getReportId().getId();
        final int submitId = submit.getId();
        final String action = "BUSY"; // NO18N
            
        // then, when you want to schedule a task
        Runnable task = new Runnable() {
            @Override
            public void run() {
                try {
                    
                    ServletRequest request = thisTask.getAsyncContext().getRequest();
                    ServletResponse response = thisTask.getAsyncContext().getResponse();
                    request.setAttribute("action", action); // NOI18N
                    request.setAttribute("reportNo", reportId); // NOI18N
                    request.setAttribute("bugNo", submitId); // NOI18N
                    request.getRequestDispatcher("/uploadredirect.jsp").forward(request, response); // NOI18N
                    LOG.log(Level.INFO, "[sbm {1}] Server busy - sending async servlet reply : {0}. Submit {1}.", new Object[] { action, String.valueOf(submitId) } ); // NOI18N
                    thisTask.getAsyncContext().complete();

                } catch (ServletException | IOException | IllegalStateException ex) {
                    // Illegal state : in case response has timed out - we still need to have a bug stored in db
                    //     although we have entered it late (after async server response http timeout in this case)
                    //     - thereofore no rethrow

                    // This is not very clean code, but there's really no check like "isComplete()" on AsyncContext API
                    // (Note: if there's one in the future, it needs to be synchronided)
                    if (!(ex instanceof IllegalStateException) || !ex.getMessage().contains("already completed")) {
                        LOG.log(Level.SEVERE, ex.getMessage(), ex);
                    }
                }
            }
        };
        
        scheduledExecutor.schedule(task, SERVER_BUSY_REPLY_IN_SECONDS, TimeUnit.SECONDS);
                        
            Integer beforeReopenId = null;
            if (duplicatesReport != null){
                beforeReopenId = duplicatesReport.getId();
            }
            BuildInfo bi = createBuildInfo(buildInfo);
            if (bi.getVcs() == BuildInfo.Vcs.ADE || bi.getVcs() == BuildInfo.Vcs.GIT) {
//                Date d = bi.getBuildDate();
                Logfile logFile = submit.getLogfileId();
                logFile.setBuildnumber(buildFromLabel(bi.getIdentificator()));
                thisTask.getTaskEntityManager().merge(logFile);
            }
            submit.setBuildId(bi);
            BugTrackerReopenChecker reopenChecker = BugTrackerProvidersFactory.getInstance().getReopenChecker(
                new BugTrackerReopenChecker() {
                    @Override
                    public boolean checkDuplicateReportForReopen(
                            Submit sbmt, ReopenCallback callback) {
                        return DbInsertion.this.checkDuplicateReportForReopen(sbmt);
                    }
                }
            );

            if (IZInsertion.isBugFiling() || IZInsertion.isTestingReopen()) {
                INSERTION_LOG.log(Level.INFO, "[sbm {0}] Checking for reopen.", String.valueOf(submitId)); // NOI18N
                boolean reopen = reopenChecker.checkDuplicateReportForReopen(submit, new ReopenCallback() {
                    @Override
                    public void handleReopen(Submit submit) {
                        DbInsertion.this.handleReopen(submit);
                    }

                    @Override
                    public LogRecord getBuildInfo() {
                        return DbInsertion.this.buildInfo;
                    }
                    
                    @Override
                    public AsyncContext getAsyncContext() {
                        return thisTask.getAsyncContext();
                    }
                    
                    @Override
                    public EntityManager getEntityManager() {
                        return thisTask.getTaskEntityManager();
                    }
                });

                if (IZInsertion.isBugFiling()) {
                    if (!reopen && submit.getReportId().isInIssuezilla()) {
                        processAsyncRequestReply(submit.getReportId(), "EXISTS");
                    }

                    if (!reopen){
                        INSERTION_LOG.log(Level.INFO, "[sbm {0}] Reopen check negative", String.valueOf(submitId)); // NOI18N
                        INSERTION_LOG.log(Level.INFO, "[sbm {0}] Requested BugDB action (insertion/comment)", String.valueOf(submitId)); // NOI18N
                        new IZInsertion(submit, preparedParams.getUserToken(), thisTask.getAsyncContext()).start();
                    } else {
                        INSERTION_LOG.log(Level.INFO, "[sbm {0}] Reopen check was positive and reopen performed by BugTrackerReopenChecker.", String.valueOf(submitId)); // NOI18N
                        data.setReportBeforeReopenId(beforeReopenId);
                    }
                }
            } else {
                // TODO: async reply that bug filing is turned off!
                processAsyncRequestReply(submit.getReportId(), "NOBUGFILING");
                INSERTION_LOG.log(Level.INFO, "[sbm {0}] Bug filing is turned off, skipping reopen check.", String.valueOf(submitId)); // NOI18N
            }
        }
        if (duplicatesReport != null) {
            data.setIssuezillaId(duplicatesReport.getIssueId());
            data.setReportId(duplicatesReport.getId());
        }
        if (!isGuest(preparedParams.getUserName())) {
            data.setUserName(preparedParams.getUserName());
        }
        time = System.currentTimeMillis() - time;
        INSERTION_LOG.log(Level.FINE, "[sbm {1}] submit db insertion finished in {0}ms", new Object[] { time, String.valueOf(submitId) } );  // NOI18N
        
    }

    private void processAsyncRequestReply(Report report, String action) {
        try {
            // Let's try to make async processing
            if (thisTask.isAsyncProcessing()) {
                INSERTION_LOG.log(Level.INFO, "[sbm {1}] Sending async servlet reply : {0}. Submit {1}", new Object[] { action, String.valueOf(submitId) } ); // NOI18N
                AsyncContext asyncContext = thisTask.getAsyncContext();
                    ServletRequest request = asyncContext.getRequest();
                    ServletResponse response = asyncContext.getResponse();
                    request.setAttribute("action", action);
                    request.setAttribute("reportNo", report.getId());
                    request.setAttribute("bugNo", report.getIssueId());
                    request.getRequestDispatcher("/uploadredirect.jsp").forward(request, response); // NOI18N
                    asyncContext.complete();
            }            
        } catch (ServletException | IOException | IllegalStateException ex) {
            // Illegal state : in case response has timed out - we still need to have a bug stored in db
            //     although we have entered it late (after async server response http timeout in this case)
            //     - thereofore no rethrow

            // This is a little hack, but there's really no check like "isComplete()" on AsyncContext API
            // (Note: if there's one in the future, it needs to be synchronided)
            if (!(ex instanceof IllegalStateException) || !ex.getMessage().contains("already completed")) {
                INSERTION_LOG.log(Level.SEVERE, ex.getMessage(), ex);
            }    
        }
    }

    @Deprecated
    private boolean checkDuplicateReportForReopen(Submit sbmt){
        if (!sbmt.getReportId().isInIssuezilla()){
            LOG.fine("Not in issuezilla");
            return false;
        }
        Integer issueId = sbmt.getReportId().getIssueId();
        Issue issue = bugReporter.getIssue(issueId);
        if (issue == null){
            LOG.log(Level.SEVERE, "[sbm {1}] Impossible to find issue for issueId = {0}", new Object[] { issueId, sbmt.getId().toString() }); // NOI18N
            return false;
        } else {
            LOG.log(Level.FINE, "[sbm {0}] Issue is not null", sbmt.getId().toString()); // NOI18N
        }
        // (FIXED || WORKSFORME || INCOMPLETE) && USER NOT GUEST
        if ((issue.isFixed()||issue.isWorksForMe())||issue.isIncomplete() && !Utils.isGuest(sbmt.getNbuserId().getName())) {
            // allow reopen even if there's no comment 
            //if (sbmt.getCommentCollection() == null || sbmt.getCommentCollection().isEmpty()){
            //    LOG.fine("No comments");
            //    return false;
            //}
            Long fixedBuildNumber = BuildQuery.findBuild(issueId); 
            if (org.netbeans.modules.exceptions.utils.Utils.isDailyBuild(sbmt.getLogfileId().getProductVersionId()) && sbmt.isDevVersion() && fixedBuildNumber != null){
                Long excBuildNumber = sbmt.getLogfileId().getBuildnumber();
                if ((excBuildNumber != null) && (Utils.isReportedAfterFixed(fixedBuildNumber, excBuildNumber))){
                    Object[] params = new Object[]{sbmt.getId(), excBuildNumber, fixedBuildNumber};
                    //cancel reopen if there's already a newer sumbit under report
                    if(!isSubmitNewest(sbmt)) {
                        LOG.log(Level.INFO, "Cancel reopen, there's newer sumbit under report #{0} that {1}", // NOI18N
                                new Object[]{sbmt.getReportId().getId(), sbmt.getId()});
                        return false;
                    }
                    if(issue.isIncomplete()){
                        //handle INCOMPLETE by direct reopen
                        LOG.log(Level.INFO, "Calling direct reopen of INCOMPLETE issue {0} excBuild: {1}; fixed build: {2}", params); // NOI18N
                        directReopen(sbmt);
                        return true;
                    } else {
                        LOG.log(Level.INFO, "Calling reopen of issue {0} excBuild: {1}; fixed build: {2}", params); // NOI18N
                        handleReopen(sbmt);
                        return true;
                    }
                }
            }
            Date date = issue.getLastResolutionChange();
            if ((date != null) && (sbmt.isBuildNewerThan(date)) && isIssueTMBeforeExcProductVersion(sbmt, issue)){
                Object[] params = new Object[]{sbmt.getId(), sbmt.getBuild(), date, sbmt.getLogfileId().getProductVersionId().getProductVersion(), issue.getTMNumber()};
                //cancel reopen if there's already a newer sumbit under report
                if(!isSubmitNewest(sbmt)) {
                    LOG.log(Level.INFO, "Cancel reopen, there's newer sumbit under report #{0} that {1}", // NOI18N
                            new Object[]{sbmt.getReportId().getId(), sbmt.getId()});
                    return false;
                }
                LOG.log(Level.INFO, "Calling reopen of issue {0} excBuild: {1}; date: {2} ; excVersion: {3}; issueTM: {4}", params); // NOI18N
                handleReopen(sbmt);
                return true;
            } else {
                LOG.log(Level.FINE, "Is not newer {0}, {1},{2}", new Object[]{date, issue.getTMNumber(), issue.getTargetMilestone()}); // NOI18N
            }
        } else {
            LOG.fine("Not fixed or is guest");
        }
        return false;
    }

    @Deprecated
    private void directReopen(Submit submit) {
        Integer issueId = submit.getReportId().getIssueId();
        String sumbitLink = "http://statistics.netbeans.org/exceptions/exception.do?id=" + submitId; // NOI18N
        String reopenComment = NbBundle.getMessage(IZInsertion.class,"direct_reopen_msg", sumbitLink); // NOI18N
        bugReporter.reopenIssue(issueId, reopenComment, preparedParams.getUserToken().getUserName(), preparedParams.getUserToken().getPasswd());
    }

    @Deprecated    
    private boolean isSubmitNewest(Submit submit){
        boolean isSubmNewest = true;
        long buildNr = submit.getBuild();
        Report r = submit.getReportId();
        r.preloadSubmitCollection(thisTask.getTaskEntityManager());
        List<Submit> submits = r.getSubmitCollection();
        for (Submit s : submits) {
            if(s.getBuild()>buildNr) {
                isSubmNewest = false;
            }
        }
        return isSubmNewest;
    }
    
    private void handleReopen(Submit submit){
        LOG.log(Level.INFO, "[sbm {0}] Trying to reopen issue exc id: {0}", submit.getId().toString()); // NOI18N
        Integer issuezillaId = submit.getReportId().getIssueId();
        EntityManager em = thisTask.getTaskEntityManager();
        Submit afterSplit = ReopenHandler.createHandler(submit).process(em, categorizer, submit.getTestName());
        em.getTransaction().commit(); // commit reopen handler changes for IZ Insertion
        em.getTransaction().begin();
        duplicatesReport = afterSplit.getReportId();
        LOG.log(Level.INFO, "[sbm {0}] Trying to report after reopen split the exception id: {0}", submit.getId().toString()); // NOI18N
        
        boolean isNewReport = !duplicatesReport.isInIssuezilla();
        
        String reopenMessage = NbBundle.getMessage(IZInsertion.class, "originally_reported", String.valueOf(issuezillaId)); // NOI18N
        if (IZInsertion.isBugFiling()) {
            if (isNewReport) {
                new IZInsertion(afterSplit, preparedParams.getUserToken(), thisTask.getAsyncContext(), null, reopenMessage).startReopenInsert();
            } else {
                new IZInsertion(submit, preparedParams.getUserToken(), thisTask.getAsyncContext()).start(); // only comment will be inserted in this case
            }
            
        }
    }

    private void insertLogFileInformation(int submitId) {
        INSERTION_LOG.log(Level.INFO, "[sbm {0}] Inserting logfile information into DB", String.valueOf(submitId));   // NOI18N
        EntityManager em = thisTask.getTaskEntityManager();
        ProductVersion productVersion = null;

        if (preparedParams.getVersion() != null) {
            productVersion = getExistProductVersion(em, preparedParams.getVersion());
        } else {
            List<ProductVersion> list = PersistenceUtils.executeQuery(em, "SELECT p FROM ProductVersion p WHERE p.productVersion is null", null); // NOI18N
            if (!list.isEmpty()) {
                productVersion = list.get(0);
            }
        }
        if (productVersion == null) {
            productVersion = new ProductVersion(PersistenceUtils.getNextId(ProductVersion.class));
            productVersion.setProductVersion(preparedParams.getVersion());
            String nbVersionName = org.netbeans.modules.exceptions.utils.Utils.getNbVersion(preparedParams.getVersion());
            Nbversion nbVersion = getExistNbVersion(em, nbVersionName);
            if (nbVersion == null) {
                nbVersion = new Nbversion(PersistenceUtils.getNextId(Nbversion.class));
                nbVersion.setVersion(nbVersionName);
                em.persist(nbVersion);
            }
            productVersion.setNbversionId(nbVersion);
            em.persist(productVersion);
        }
        Long changeset = parseBuildInfo(buildInfo);
        Logfile logFile = thisTask.getLogInfo();
        assert logFile != null;
        logFile.setProductVersionId(productVersion);
        logFile.setChangeset(changeset);
        logFile.setBuildnumber(preparedParams.getBuildNumber());
        em.persist(logFile);
        em.merge(logFile);
    }

    private Nbuser prepareNBUser(EntityManager em){
        Nbuser nbuser = getExistUser(em, preparedParams.getUserName());
        if (nbuser == null) {//Adding NbUser into DB
            nbuser = createUser(em, preparedParams.getUserName());
        }
        return nbuser;
    }

    private Component doComponentChecking(EntityManager em, Submit exc, Report report){
        // get component from comment
        Component comp = getComponentFromComment(preparedParams.getComment());
        boolean isDCS = report.getReportType().equals("DCS");
        if (isDCS) {
            // we don't use stacktrace mappings for DCS, it is determined from data sent in a report
            
            String component = exc.getAdditionalInfo("productName");
            String subcomponent = exc.getAdditionalInfo("componentName");

            if (component == null || subcomponent == null ||
                component.isEmpty() || subcomponent.isEmpty()) {
                return DEFAULT_DCS_COMPONENT;
            } else {
                Component retValue = new Component(component, subcomponent);
                retValue.setProduct(DCS_PRODUCT);
                return retValue;
            }
        }
        // use unknown component for STACK OVERFLOW ISSUE 112156
        if ((comp == null) && thrown.getMessage().contains("StackOverflowError")) { // NOI18N
            comp = DEFAULT_JDEV_COMPONENT;
        }else if (comp == null) {
            // TODO 
            // perform refresh of innocents and sourcejar mappings
            Matcher.getDefault().reload();
            comp = Matcher.getDefault().match(em, thrown);
        }
        
        return comp;
    }

    private Nbuser createUser(EntityManager em, String userName) {
        Nbuser nbuser = new Nbuser(PersistenceUtils.getNextId(Nbuser.class));
        nbuser.setName(userName);
        em.persist(nbuser);
        return nbuser;
    }

    protected static Component getComponentFromComment(String comment) {
        int startIndex = comment.indexOf("[");
        int slashIndex = comment.indexOf("/");
        int endIndex = comment.indexOf("]");
        int newLineIndex = comment.indexOf("\n");
        if (newLineIndex == -1) {
            newLineIndex = comment.length();
        }
        if ((startIndex == 0) && (startIndex < slashIndex) &&
                (slashIndex < endIndex) && (endIndex < newLineIndex)) {
            String comp = comment.substring(1, slashIndex);
            String subcomp = comment.substring(slashIndex + 1, endIndex);

            Components realComps = PersistenceUtils.getInstance().getComponents();
            if (realComps.getComponentsSet().contains(comp) && realComps.getSubcomponentsSet(comp).contains(subcomp)) {
                return new Component(comp, subcomp);
            }
        }
        return null;
    }

    private Submit insertSlownessData(SlownessChecker checker) {
        EntityManager em = thisTask.getTaskEntityManager();

        Nbuser user = prepareNBUser(em);
        Slowness sn = new Slowness(submitId);
        sn.setSummary(preparedParams.getSummary());
        sn.setLogfileId(thisTask.getLogInfo());
        sn.setVmId(getVm(em, preparedParams.getVm()));
        sn.setOperatingsystemId(getOperatingsystem(em, preparedParams.getOs()));
        sn.setNbuserId(user);
        sn.setReportdate(new Date());
        sn.setLatestAction(latestAction);
        sn.setActionTime(slownessTime);
        sn.setSlownessType(slownessType);
        sn.setSlowWindowsIO(checker.hasSlowWindowsIO());
        String fullMethodName = checker.getSuspiciousMethodName();
        Method method = getExistingMethod(em, fullMethodName);
        if (method == null){
            method = new Method();
            method.setId(PersistenceUtils.getNextId(Method.class));
            method.setName(fullMethodName);
            em.persist(method);
        }
        sn.setSuspiciousMethod(method);
        Report report = prepareReport(em, sn);

        // get component from comment
        Component comp = getComponentFromComment(preparedParams.getComment());
        if (comp == null){
            comp = checker.getComponentForSlowness();
        }
        if (comp == null || comp.equals(Component.UNKNOWN)) {
            comp = Component.UNKNOWN_SLOWNESS;
        }
        report.setReportType(preparedParams.getReportType());
        report.setComponent(comp);
        sn.setReportId(report);
        em.persist(sn);
        em.merge(report);
        sn = mergeComment(em, sn, user);

        em.getTransaction().commit();
        em.getTransaction().begin();
        return sn;
    }

    private <T extends Submit> T mergeComment(EntityManager em, T submit, Nbuser user){
        // adding comment
        if (!preparedParams.getComment().isEmpty()){
            Comment comment = new Comment();
            comment.generateId();
            comment.setCommentDate(new Date());
            comment.setSubmitId(submit);
            comment.setNbuserId(user);
            comment.setComment(preparedParams.getComment());
            em.persist(comment);
            Collection<Comment> commentList = new ArrayList<Comment>();
            commentList.add(comment);
            submit.setCommentCollection(commentList);
            return em.merge(submit);
        }else{
            return submit;
        }
    }

    private Report prepareReport(EntityManager em, Submit submit){
        if (duplicatesReport != null) {
            duplicatesReport.preloadSubmitCollection(em);
            duplicatesReport.getSubmitCollection().add(submit);
            return duplicatesReport;
        } else {
            Report report = createReport(em);
            List<Submit> list = new ArrayList<Submit>();
            list.add(submit);
            report.setSubmitCollection(list);
            return report;
        }
    }

    private void addCause(Throwable thrownCause, ArrayList<StackTraceElement> newStack) {
        if (thrownCause.getCause() != null) {
            addCause(thrownCause.getCause(), newStack);
        }
        for (StackTraceElement ste : thrownCause.getStackTrace()) {
            newStack.add(ste);
        }            
    }
    
    private StackTraceElement[] getMergedStackTrace(Throwable thrown) {
        ArrayList<StackTraceElement> newStack = new ArrayList<StackTraceElement>();
        addCause(thrown, newStack);
        return newStack.toArray(new StackTraceElement[newStack.size()]);
    }

    // we're storing the stacktrace in so-called "merged" form: on each level, the stacktrace contains also all lines from their children
    private Throwable mergeThrowableHierarchicaly(Throwable thrown) {
        
        if (thrown.getCause() == null) {
            return thrown;
        }
        Throwable toReturn = new Throwable(thrown.getMessage(), mergeThrowableHierarchicaly(thrown.getCause()));
        toReturn.setStackTrace(getMergedStackTrace(thrown));
        
        return toReturn;
    }
            
    // See http://aseng-wiki.us.oracle.com/asengwiki/display/ASDevJDeveloper/ER+Matching+Engine for some code clarifications
    private Submit insertException() {
        INSERTION_LOG.log(Level.INFO, "[sbm {0}] Inserting exception into DB {0}", Integer.toString(submitId));   // NOI18N
        final EntityManager em = thisTask.getTaskEntityManager();

        int apiMark = preparedParams.getApiMark();

        Throwable mergedThrown = mergeThrowableHierarchicaly(thrown);
        
        // adding stacktrace
        recursionDepth = 0; //reset recursion counter
        Stacktrace stacktrace = addStackTrace(em, mergedThrown, apiMark);
        Nbuser nbuser = prepareNBUser(em);
        
        org.netbeans.modules.exceptions.entity.Exceptions exc = new org.netbeans.modules.exceptions.entity.Exceptions();
        Logfile logFile = thisTask.getLogInfo();
        exc.setId(submitId);
        exc.setSummary(preparedParams.getSummary());
        exc.setStacktrace(stacktrace);
        exc.setLogfileId(logFile);
        exc.setVmId(getVm(em, preparedParams.getVm()));
        exc.setOperatingsystemId(getOperatingsystem(em, preparedParams.getOs()));
        exc.setBuild(logFile.getBuildnumber());
        exc.setNbuserId(nbuser);
        exc.setTestName(preparedParams.getTestName());
//        exc.setTestName(preparedParams.getTestName());
//        if (preparedParams.getTestName() != null) {
//            exc.setSource(org.netbeans.modules.exceptions.entity.Exceptions.Source.SRG);
//        } else {
//            exc.setSource(org.netbeans.modules.exceptions.entity.Exceptions.Source.CLIENT);
//        }
        exc.setTags(preparedParams.getTags());
        Report report = prepareReport(em, exc);
        report.setReportType(preparedParams.getReportType());
        Component comp = doComponentChecking(em, exc, report);
        if (comp == null) {
            comp = DEFAULT_JDEV_COMPONENT;
        }
        report.setComponent(comp);
        exc.setReportId(report);
        exc.setReportdate(new Date());
        em.persist(exc);
        em.merge(report);
        stacktrace.setExceptions(exc);
        em.merge(stacktrace);

        exc = mergeComment(em, exc, nbuser);

        //add hashCode
        Hashcodes hashcode = new Hashcodes(submitId);
        hashcode.setCode(countHash(mergedThrown.getStackTrace(), apiMark));
        hashcode.setExceptions(exc);
        em.persist(hashcode);
        exc.setHashcodes(hashcode);
        em.merge(exc);

        em.getTransaction().commit();
        em.getTransaction().begin();
        return exc;
    }

    private Stacktrace addStackTrace(EntityManager em, Throwable thrown, int apiMark) {
        Stacktrace stacktrace = new Stacktrace();
        Line line;
        LinePK linePK;
        Method method;
        Jarfile jarFile;

        int stackTraceId = PersistenceUtils.getNextId(Stacktrace.class);
        Stacktrace cause = null;

        if (thrown.getCause() != null && recursionDepth < 10) {
            recursionDepth++;
            cause = addStackTrace(em, thrown.getCause(), 0);//add recursively
            // apiMark applies only to the toplevel exception, therefore it's 0 on recursive cause calls
        }
        stacktrace.setId(stackTraceId);
        String message = thrown.getMessage();
        String stackClass = message;
        if (message != null) {
            if (message.length() > MAXMESSAGELENGTH) {
                message = message.substring(0, MAXMESSAGELENGTH);
            }
            stacktrace.setMessage(message);
            stackClass = FilteringUtils.getClassOfMessage(message);
            stacktrace.setClass1(stackClass);
        }
        if (cause != null) {
            stacktrace.setAnnotation(cause);
        }
        //add stack trace lines
        em.persist(stacktrace);
        if (cause != null) {
            cause.setStacktrace(stacktrace);
            em.merge(cause);
        }
        Integer lineNumber;
        String linesConcatenation = null;
        Collection<Line> lineCollection = new ArrayList<Line>(thrown.getStackTrace().length);
        boolean isExcluded = true;
        boolean justLeftFramework = false;
        for (int i = 0; i < thrown.getStackTrace().length; i++) {
            StackTraceElement element = thrown.getStackTrace()[i];
            String jarFileName = element.getFileName();
            if ("".equals(jarFileName)) {
                jarFileName = null;
            }
            if (isExcluded && !categorizer.isPackageExcluded(element.getClassName())) {
                isExcluded = false;
                justLeftFramework = true;
            }
            if (linesConcatenation == null || i == apiMark || justLeftFramework) {
                // we start counting lines concatenation from apiMark as we
                // want to ignore lines below apiMark for future matching
                linesConcatenation = Method.getMethodNameFromSTEWithoutNumbers(element);
                justLeftFramework = false;                
            } else {
                linesConcatenation = linesConcatenation.concat(Method.getMethodNameFromSTEWithoutNumbers(element));
            }
            String fullMethodName = Method.getMethodNameFromSTE(element);
            method = getExistingMethod(em, fullMethodName);
            if (method == null) {//adding method into table
                method = new Method();
                method.setId(PersistenceUtils.getNextId(Method.class));
                method.setName(fullMethodName);
                if (jarFileName != null) {
                    method.setSourcejar(getSourceJar(em, jarFileName));
                }
                em.persist(method);
            } else {
                if ((method.getSourcejar() == null) && (jarFileName != null)) {
                    method.setSourcejar(getSourceJar(em, jarFileName));
                    method = em.merge(method);
                }
            }
            if (jarFileName == null) {
                jarFileName = UNKNOWN;
            }
            jarFile = getExistingFile(em, jarFileName);
            if (jarFile == null) {
                jarFile = new Jarfile(PersistenceUtils.getNextId(Jarfile.class));
                jarFile.setName(jarFileName);
                em.persist(jarFile);
            }
            if (element.getLineNumber() >= 0) {
                lineNumber = element.getLineNumber();
            } else {
                lineNumber = 0;
            }
            line = new Line();
            linePK = new LinePK();
            linePK.setLinenumber(lineNumber);
            linePK.setMethodId(method.getId());
            linePK.setStacktraceId(stackTraceId);
            linePK.setLineOrder(i);
            line.setStacktrace(stacktrace);
            line.setMethod(method);
            line.setLineHashcode(linesConcatenation.hashCode());
            lineCollection.add(line);
            line.setJarfileId(jarFile);
            line.setLinePK(linePK);
            line.setGeneratecomponent(0);
            em.persist(line);
        }
        stacktrace.setLineCollection(lineCollection);
        stacktrace = em.merge(stacktrace);
        return stacktrace;    
    }

    private static Sourcejar getSourceJar(EntityManager em, String jarFileName) {
        String[] parts = jarFileName.split("/");
        String jarName = parts[parts.length - 1];
        Sourcejar sjar = (Sourcejar) PersistenceUtils.getExist(em, "Sourcejar.findByName", jarName);
        if (sjar == null) {
            sjar = new Sourcejar(PersistenceUtils.getNextId(Sourcejar.class));
            sjar.setName(jarName);
            em.persist(sjar);
        }
        return sjar;
    }

    private Report createReport(EntityManager em) {
        Report report = new Report();
        report.setId(PersistenceUtils.getNextId(Report.class));
        report.setThrottled(0);
        em.persist(report);
        return report;
    }

    private Operatingsystem getOperatingsystem(EntityManager em, String osName) {
        String [] s = osName.split(",");
        String name = "";
        String version = "";
        String arch = "";
        if (s.length == 3) {
            name = s[0].trim();
            version = s[1].trim();
            arch = s[2].trim();
        } else {
            name = osName;
        }
        Operatingsystem os = Operatingsystem.getExists(em, name, version, arch);
        if (os == null) {
            os = new Operatingsystem();
            os.setId(PersistenceUtils.getNextId(Operatingsystem.class));
            os.setName(name);
            os.setVersion(version);
            os.setArch(arch);
            em.persist(os);
        }
        return os;
    }
    private Vm getVm(EntityManager em, String vmName) {
        Vm vm = (Vm) PersistenceUtils.getExist(em, "Vm.findByName", vmName); // NOI18N
        if (vm == null) {
            vm = new Vm();
            vm.setId(PersistenceUtils.getNextId(Vm.class));
            vm.setName(vmName);
            vm.setShortName(getShortVM(vmName));
            em.persist(vm);
        }
        return vm;
    }

    private String getShortVM(String vm) {
        String result = "";
        if (vm != null) {
            for (int i = 6; i < 10; i++) {
                String v = "1." + i;
                if (vm.contains(v + ".0")) {
                    result = v;
                    break;
                }
            }
        }
        return result;
    }

    static Long parseBuildInfo(LogRecord buildInfoRec) {
        if (buildInfoRec != null){
            Object[] infoParams = buildInfoRec.getParameters();
            if ((infoParams != null) && (infoParams.length >= CHANGESET_IDX + 1)){
                String param = infoParams[CHANGESET_IDX].toString();
                if (param.length() == 0){
                    return null;
                }
                int space = param.indexOf(' ');
                if (space != -1){
                    param = param.substring(0, space);
                }
                param = param.replaceAll("\\+", "");
                try {
                    return Long.parseLong(param, CHANGESET_RADIX);
                }catch (NumberFormatException nfe){
                    LOG.log(Level.WARNING, "Wrong changeset format: {0}", infoParams[CHANGESET_IDX].toString()); // NOI18N
                }
            }
        }
        return null;
    }

    private Jarfile getExistingFile(EntityManager em, String fileName) {
        return (Jarfile) PersistenceUtils.getExist(em, "Jarfile.findByName", fileName);// NOI18N
    }

    private Method getExistingMethod(EntityManager em, String methodName) {
        return (Method) PersistenceUtils.getExist(em, "Method.findByName", methodName); // NOI18N
    }

    private Nbuser getExistUser(EntityManager em, String userName) {
        return (Nbuser) PersistenceUtils.getExist(em, "Nbuser.findByName", userName);// NOI18N
    }

    private ProductVersion getExistProductVersion(EntityManager em, String versionName) {
        return (ProductVersion) PersistenceUtils.getExist(em, "ProductVersion.findByProductVersion", versionName);// NOI18N
    }

    private Nbversion getExistNbVersion(EntityManager em, String versionName) {
        return (Nbversion) PersistenceUtils.getExist(em, "Nbversion.findByVersion", versionName);// NOI18N
    }

    private BuildInfo createBuildInfo(LogRecord buildInfo) {
        if (buildInfo == null || buildInfo.getParameters() == null) {
            return getNABuildInfo(thisTask.getTaskEntityManager());
        }
        BuildInfo result;
        Object[] params = buildInfo.getParameters();
        Vcs vcs = Vcs.fromString((String) params[0]);
        String identificator = null;
        String version = "";
        switch (vcs) {
            case ADE:
            case GIT:
            case HG: identificator = (String) params[1];
                     version = (String) params[2];
                break;
            default: if (params.length > 5) {
                         identificator = (String) params[5];
                     }
        }
        Query nq = thisTask.getTaskEntityManager().createNamedQuery("BuildInfo.findByIdentificator"); // NOI18N
        nq.setParameter("identificator", identificator);
        List<BuildInfo> r = nq.getResultList();
        if (r.isEmpty()) {
            result = new BuildInfo();
            result.setIdentificator(identificator);
            result.setVersion(version);
            result.setVcs(vcs);
            if (identificator != null) {
                result.setBuildDate(dateFromLabel(identificator));
            }
            result.setId(PersistenceUtils.getNextId(BuildInfo.class));
            thisTask.getTaskEntityManager().persist(result);
        } else {
            result = r.iterator().next();
        }
        return result;
    }
    
    
    private Date dateFromLabel(String label) {
        java.util.regex.Matcher m = LABEL_PATTERN.matcher(label);
        if (m.matches()) {
            try {
                return labelDateFormat.parse(m.group(4));
            } catch (ParseException ex) {
                LOG.log(Level.SEVERE, "Not able to parse label {0}", label); // NOI18N
            }
        }
        return null;
    }
    
    private long buildFromLabel(String label) {
        long result = 0;
        java.util.regex.Matcher m = LABEL_PATTERN.matcher(label);
        if (m.matches()) {
            try {
                result = Long.valueOf(m.group(4)) * 10000;
                result += Integer.valueOf(m.group(5));
                
            } catch (NumberFormatException ex) {
                LOG.log(Level.SEVERE, "Not able to parse label {0}", label); // NOI18N
            }
        }
        return result;
    }
    
    private BuildInfo getNABuildInfo(EntityManager em) {
        if (naBuildInfo == null) {
            List<BuildInfo> r = em.createQuery("SELECT b FROM BuildInfo b WHERE b.identificator = 'N/A'").getResultList(); // NOI18N
            if (r.isEmpty()) {
                BuildInfo nbBuildInfo = new BuildInfo();
                nbBuildInfo.setVersion("N/A");
                nbBuildInfo.setVcs(BuildInfo.Vcs.NONE);
                nbBuildInfo.setId(PersistenceUtils.getNextId(BuildInfo.class));
                em.persist(nbBuildInfo);
                naBuildInfo = nbBuildInfo;
            } else {
                naBuildInfo = r.iterator().next();
            }
        }
        return naBuildInfo;
    }

    public static class InsertThrottledException extends RuntimeException {

        public InsertThrottledException() {
        }
    }

}
