/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2006 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.server.uihandler;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.netbeans.server.uihandler.api.Authenticator;
import org.netbeans.server.uihandler.api.Authenticator.AuthToken;
import org.netbeans.server.uihandler.api.bugs.BugReporter;

/**
 *
 * @author Jindrich Sedek
 */
public class PreparedParams {

    private static final Logger LOG = Logger.getLogger(PreparedParams.class.getName());
    
    private static final int MAXSUMMARY = 200;
    private static final int MAXCOMMENT = 1000;
    
    private static final String REPORT_TYPE_KEY = "reportType";
    private static final String OS_KEY = "os";
    private static final String VM_KEY = "vm";
    private static final String VERSION_KEY = "version";
    private static final String USER_NAME_KEY = "user";
    private static final String SUMMARY_KEY = "title";
    private static final String COMMENT_KEY = "comment";
    private static final String SRG_TESTCASE_KEY = "testCase";
    private static final String API_MARK_KEY = "apiMark";
    private static final String SUN_DESKTOP_KEY = "sunDesktop";
    private static final String TAGS_KEY = "tags";

    private final String reportType;
    private final String os;
    private final String vm;
    private final String version;
    private final String summary;
    private final String comment;
    private final Authenticator.AuthToken userToken;
    private final String userName;
    private final String srgTestName;
    private final String apiMark;
    private final String sunDesktop;
    private final String tags;
    private final Map<String, String> additionalParams;

    private PreparedParams(String reportType, String os, String vm, String version, String summary, 
            String comment, String userName, String srgTestcase, String apiMark,
            String sunDesktop, String tags, Throwable thrown, Map<String, String> additionalParams) {
        this.reportType = reportType;
        this.os = os;
        this.vm = vm;
        this.version = version;
        
        if (summary != null) {
            if ((summary.length() == 0) && (thrown != null)) {
                summary = thrown.getMessage();
            }
            if (summary.length() > MAXSUMMARY) {
                summary = summary.substring(0, MAXSUMMARY);
            }
        }
        this.summary = summary;
         
        if (comment != null) {
            if (comment.length() > MAXCOMMENT) {
                comment = comment.substring(0, MAXCOMMENT);
            }
        }
        
        if (comment == null) {
            this.comment = "";
        } else {
            this.comment = comment;
        }
        if (userName == null || userName.isEmpty()) {
            userName = BugReporter.GUEST_USER;
        }
        this.userName = userName;
        this.userToken = null;
        this.srgTestName = srgTestcase;
        this.apiMark = apiMark;
        this.sunDesktop = sunDesktop;
        this.tags = tags;
        this.additionalParams = additionalParams;
    }

    public static PreparedParams prepare(Object[] oParams, Throwable thrown) {
        
        Map<String,String> params = new LinkedHashMap<> ();
        
        for (Object obj : oParams) {
            String param = obj.toString();
            String[] nameValuesArray = param.split("=", 2);
            if (nameValuesArray.length < 2) {
                continue;
            }
            String name = nameValuesArray[0];
            String value = nameValuesArray[1];
            params.put(name, value);
        }

        return new PreparedParams(params.remove(REPORT_TYPE_KEY), params.remove(OS_KEY), params.remove(VM_KEY), 
            params.remove(VERSION_KEY), params.remove(SUMMARY_KEY), params.remove(COMMENT_KEY),
            params.remove(USER_NAME_KEY), params.remove(SRG_TESTCASE_KEY), params.remove(API_MARK_KEY),
            params.remove(SUN_DESKTOP_KEY), params.remove(TAGS_KEY), thrown, params);
    }
        
    public String getReportType() {
        return reportType;
    }

    public String getComment() {
        return comment;
    }

    public String getOs() {
        return os;
    }

    public String getSummary() {
        return summary;
    }

    public AuthToken getUserToken() {
        return userToken;
    }

    public String getUserName() {
        if (userToken != null) {
            return userToken.getUserName();
        } else {
            return userName;/*BugReporter.GUEST_USER;*/
        }
    }

    public String getVersion() {
        return version;
    }

    public String getVm() {
        return vm;
    }
    
    public String getTestName() {
        return srgTestName;
    }

    public Long getBuildNumber(){
        if (version != null){
            String buildNumber = org.netbeans.modules.exceptions.utils.Utils.getBuildNumber(version);
            buildNumber = org.netbeans.modules.exceptions.utils.Utils.getCustomBuildFormat(buildNumber);
            if ((buildNumber != null) && (buildNumber.length() > 0)) {
                return Long.valueOf(buildNumber);
            }
        }
        return null;
    }
    
    public int getApiMark() {
        try {
            return Integer.parseInt(apiMark);
        } catch (NumberFormatException e) {
            LOG.log(Level.WARNING, "Error parsing apiMark. Not a number : apiMark = \"{0}\"", apiMark);
            return 0;
            
        }
    }
    
    public String getSunDesktop() {
        return sunDesktop;
    }
    
    public String getTags() {
        return tags;
    }
    
    public Map<String, String> getAdditionalParams() {
        return additionalParams;
    }
}
