/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2007 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 */

package org.foo.wavutils;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.font.FontRenderContext;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.awt.geom.Line2D;
import java.io.IOException;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.util.ArrayList;
import java.util.List;
import javax.sound.sampled.AudioFormat;
import javax.swing.JPanel;
import org.openide.ErrorManager;

/**
 *
 * @author runger
 */
public class FFTGraph extends JPanel {

    //////////////////// CONSTRUCTORS /////////////////////////

    public FFTGraph() {
        setBackground(Color.black);
    }

    //////////////////// METHODS //////////////////////////////

    public void clearGraph()
    {
        freqData = null;
        maxFreq = 0f;
        m_sError = null;
        repaint();
    }

    public void createGraph( WrappedAudioInputStream audioInputStream )
    {
        if( audioInputStream != null && audioInputStream.getLength() > 0)
        {
            format = audioInputStream.getFormat();
            byte[] audioBytes = null;
            try
            {
                audioBytes = new byte[audioInputStream.getLength()];
                audioInputStream.read(audioBytes);
            }
            catch( Exception ex )
            {
                ErrorManager.getDefault().notify(ex);
                return;
            }

            createGraph(audioBytes);

            try
            {
                audioInputStream.reset();
            }
            catch( IOException ex )
            {
                ErrorManager.getDefault().notify(ex);
            }
        }
        repaint();
    }

    private void createGraph(byte[] audioBytes)
    {
        freqData = null;  // clear the old FFT data
        maxFreq = 0f;
        float[] samples;

        if (format.getSampleSizeInBits() == 16)
        {
             int iLengthInSamples = audioBytes.length / 2;
             int bits = (int)Math.ceil(Math.log(iLengthInSamples) / Math.log(2));
             samples = new float[(int)Math.pow(2, bits)];

             if (format.isBigEndian())
             {
                 for (int i = 0; i < iLengthInSamples; i++)
                 {
                     /* First byte is MSB (high order) */
                     int MSB = audioBytes[2*i];
                     /* Second byte is LSB (low order) */
                     int LSB = audioBytes[2*i+1];
                     samples[i] = MSB << 8 | (255 & LSB);
                 }
             }
             else
             {
                 for (int i = 0; i < iLengthInSamples; i++)
                 {
                     /* First byte is LSB (low order) */
                     int LSB = audioBytes[2*i];
                     /* Second byte is MSB (high order) */
                     int MSB = audioBytes[2*i+1];
                     samples[i] = MSB << 8 | (255 & LSB);
                 }
             }
        }
        else if (format.getSampleSizeInBits() == 8)
        {
             int iLengthInSamples = audioBytes.length;
             int bits = (int)Math.ceil(Math.log(iLengthInSamples) / Math.log(2));
             samples = new float[(int)Math.pow(2, bits)];

             if (format.getEncoding().toString().startsWith("PCM_SIGN"))
             {
                 for (int i = 0; i < audioBytes.length; i++)
                 {
                     samples[i] = audioBytes[i];
                 }
             }
             else
             {
                 for (int i = 0; i < audioBytes.length; i++)
                 {
                     samples[i] = audioBytes[i] - 128;
                 }
             }
        }
        else
        {
            samples = null;
            assert false : "Sample size is not 8 or 16!";
        }
        
        // perform FFT on the samples
        
        // an "empty" wav file has 1 sample
        if (samples.length < 2)
            freqData = new float[] { 0 };
        else
            freqData = new FastFourierTransform().fftMag(samples);
        
        for (float freq : freqData)
        {
            maxFreq = Math.max(maxFreq, freq);
        }
    }

    public String getErrorString()
    {
        return m_sError;
    }

    public void paint(Graphics g)
    {
        int INFOPAD = 15;
        Dimension d = getSize();
        int w = d.width;
        int h = d.height - INFOPAD;
        
        List<Line2D> lines = new ArrayList<Line2D>();
        if (freqData != null)
        {
            for (int i=0; i < freqData.length; i++)
            {
                float freq = freqData[i];
                float x = (float)w * i / freqData.length;
                float y = (freq * h / maxFreq);
                lines.add(new Line2D.Float(x, h, x, h-y));
            }
        }

        Graphics2D g2 = (Graphics2D) g;
        g2.setBackground(getBackground());
        g2.clearRect(0, 0, w, h);
        g2.setColor(Color.white);
        g2.fillRect(0, h, w, INFOPAD);

        if (m_sError != null)
        {
            // Write out an error string instead of the wave form

            g2.setColor(s_jfcBlue);
            g2.setFont(new Font("serif", Font.BOLD, 18));
            g2.drawString("ERROR", 5, 20);
            AttributedString as = new AttributedString(m_sError);
            as.addAttribute(TextAttribute.FONT, s_font12, 0, m_sError.length());
            AttributedCharacterIterator aci = as.getIterator();
            FontRenderContext frc = g2.getFontRenderContext();
            LineBreakMeasurer lbm = new LineBreakMeasurer(aci, frc);
            float x = 5, y = 25;
            lbm.setPosition(0);
            while (lbm.getPosition() < m_sError.length()) {
                TextLayout tl = lbm.nextLayout(w-x-5);
                if (!tl.isLeftToRight()) {
                    x = w - tl.getAdvance();
                }
                tl.draw(g2, x, y += tl.getAscent());
                y += tl.getDescent() + tl.getLeading();
            }
        }
        else
        {
            // write the length and position at the bottom

            g2.setColor(Color.black);
            g2.setFont(s_font12);
            g2.drawString("Fast Fourier Transform", 3, h + INFOPAD -4);

            if (!lines.isEmpty()) {
                // .. render sampling graph ..
                g2.setColor(s_jfcBlue);
                for (int i = 1; i < lines.size(); i++) {
                    g2.draw((Line2D) lines.get(i));
                }
            }

        }
    }

    //////////////////// INSTANCE VARIABLES ////////////////////

    private String m_sError;
    private float[] freqData;
    private float maxFreq;
    private AudioFormat format;
    private double m_dDuration;

    //////////////////// CONSTANTS //////////////////////////////
    private static final Font s_font12 = new Font("serif", Font.PLAIN, 12);
    private static final Color s_jfcBlue = new Color(204, 204, 255);
    private static final Color s_pink = new Color(255, 175, 175);

}
