/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2008 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.cards.api;

import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * An AddPolicy is a rule that determines whether or not cards can
 * be added to a pile. Multiple AddPolicies may be composed together
 * into a single composite AddPolicy that defines all the addition
 * rules for a Pile.
 *
 * @author Tim Boudreau
 */
public abstract class AddPolicy {

    public abstract boolean canAdd(Card card, Pile pile);
    public static final AddPolicy MATCH_SUIT = new MatchSuit();
    public static final AddPolicy ALWAYS_ADD = new DefaultAddPolicy();
    public static final AddPolicy ASCENDING = new Directional(true);
    public static final AddPolicy DESCENDING = new Directional(false);
    public static final AddPolicy RED_BLACK = new RedBlack();
    public static final AddPolicy ADJACENT = new Adjacent();
    public static final AddPolicy NEVER_ADD = new DefaultAddPolicy(false);
    public static final AddPolicy ORDERED_ASCENDING =
            createAddPolicy(ADJACENT, ASCENDING);
    public static final AddPolicy ORDERED_DESCENDING =
            createAddPolicy(ADJACENT, DESCENDING);
    public static final AddPolicy RED_BLACK_ORDERED_DESCENDING =
            createAddPolicy(ADJACENT, DESCENDING, RED_BLACK);
    public static final AddPolicy RED_BLACK_ORDERED_ASCENDING =
            createAddPolicy(ADJACENT, ASCENDING, RED_BLACK);
    public static final AddPolicy ACE_LOW = new AceLow();
    public static final AddPolicy ACE_MUST_START = new AceMustStart();
    public static final AddPolicy ACE_HIGH = new AceHigh();
    private static final Logger logger = Logger.getLogger("org.netbeans.examples.cards.api.Pile");

    static boolean log = logger.isLoggable(Level.FINE);

    private static final class Adjacent extends AddPolicy {

        public boolean canAdd(Card card, Pile pile) {
            if (pile.isEmpty()) {
                return true;
            } else {
                Rank rank = card.identity();
                Rank pileTopRank = pile.peekTopCard().identity();
                int dist = rank.compareTo(pileTopRank);
                boolean result = dist == 1 || dist == -1;
                return result;
            }
        }
    }

    private static final class AceMustStart extends AddPolicy {

        public boolean canAdd(Card card, Pile pile) {
            boolean result = !pile.isEmpty();
            if (!result) {
                result = card.identity() == Rank.ACE;
            }
            return result;
        }
    }

    private static final class AceHigh extends AddPolicy {

        public boolean canAdd(Card card, Pile pile) {
            if (pile.isEmpty()) {
                return true;
            } else {
                Rank cardRank = card.identity();
                Rank topCardRank = pile.peekTopCard().identity();
                if (cardRank == Rank.ACE) {
                    return topCardRank == Rank.KING;
                } else if (topCardRank == Rank.ACE) {
                    return cardRank == Rank.KING;
                } else {
                    return true;
                }
            }
        }
    }

    private static final class AceLow extends AddPolicy {

        public boolean canAdd(Card card, Pile pile) {
            boolean result;
            if (pile.isEmpty()) {
                result = card.identity() == Rank.ACE;
            } else {
                Rank cardRank = card.identity();
                Rank topCardRank = pile.peekTopCard().identity();
                if (cardRank == Rank.ACE) {
                    result = topCardRank == Rank.TWO;
                } else if (topCardRank == Rank.ACE) {
                    result = cardRank == Rank.TWO;
                } else {
                    result = true;
                }
            }
            return result;
        }
    }

    private static final class RedBlack extends AddPolicy {

        public boolean canAdd(Card card, Pile pile) {
            boolean result;
            if (pile.isEmpty()) {
                result = true;
            } else {
                result = card.suit().getColor() !=
                        pile.peekTopCard().suit().getColor();
            }
            return result;
        }
    }

    public static AddPolicy createAddPolicy(AddPolicy... toCompose) {
        if (toCompose.length == 0) {
            throw new IllegalArgumentException("No arguments");
        }
        if (toCompose.length == 1) {
            return toCompose[0];
        }
        return new CompositeAddPolicy(toCompose);
    }

    public static AddPolicy or(AddPolicy... toOr) {
        if (toOr.length <= 1) {
            throw new IllegalArgumentException("Must or at least two things");
        }
        return new Or(toOr);
    }

    public static AddPolicy not(AddPolicy... toNot) {
        return new Not(toNot);
    }

    private static final class CompositeAddPolicy extends AddPolicy {

        private final AddPolicy[] others;

        public CompositeAddPolicy(AddPolicy... others) {
            this.others = others;
        }

        public boolean canAdd(Card card, Pile pile) {
            boolean result = true;
            for (int i = 0; i < others.length; i++) {
                result &= others[i].canAdd(card, pile);
                if (!result) {
                    break;
                }
            }
            return result;
        }
    }

    public static final class SuitAddPolicy extends AddPolicy {

        private Suits suit;

        public SuitAddPolicy(Suits suit) {
            this.suit = suit;
        }

        public boolean canAdd(Card card, Pile pile) {
            return suit == card.suit();
        }
    }

    private static final class DefaultAddPolicy extends AddPolicy {

        private final boolean val;

        public DefaultAddPolicy(boolean val) {
            this.val = val;
        }

        public DefaultAddPolicy() {
            this(true);
        }

        public boolean canAdd(Card card, Pile pile) {
            return val;
        }
    }

    private static final class Directional extends AddPolicy {

        private boolean ascending;

        Directional(boolean ascending) {
            this.ascending = ascending;
        }

        public boolean canAdd(Card card, Pile pile) {
            boolean result;
            if (pile.isEmpty()) {
                result = true;
            } else {
                Rank rank = card.identity();
                Rank pileTopRank = pile.peekTopCard().identity();
                int dist = rank.compareTo(pileTopRank);
                result = ascending ? dist > 0 : dist < 0;
            }
            return result;
        }

        @Override
        public String toString() {
            return ascending ? "Ascending" : "Descending";
        }
    }

    private static final class Or extends AddPolicy {

        private final AddPolicy[] others;

        public Or(AddPolicy... others) {
            this.others = others;
        }

        public boolean canAdd(Card card, Pile pile) {
            boolean result = false;
            for (int i = 0; i < others.length; i++) {
                result |= others[i].canAdd(card, pile);
                if (result) {
                    break;
                }
            }
            return result;
        }
    }

    private static final class Not extends AddPolicy {

        private final AddPolicy[] others;

        public Not(AddPolicy... others) {
            this.others = others;
        }

        public boolean canAdd(Card card, Pile pile) {
            boolean result = true;
            for (int i = 0; i < others.length; i++) {
                result &= !others[i].canAdd(card, pile);
                if (!result) {
                    break;
                }
            }
            return result;
        }
    }

    private static final class MatchSuit extends AddPolicy {

        public boolean canAdd(Card card, Pile pile) {
            if (!pile.isEmpty()) {
                Suits s = pile.peekTopCard().getSuit();
                return s == card.getSuit();
            } else {
                return true;
            }
        }
    }

    public static final class DifferentRuleWhenEmpty extends AddPolicy {

        private final AddPolicy whenEmptyRule;
        private final AddPolicy whenNotEmptyRule;

        public DifferentRuleWhenEmpty(AddPolicy whenEmptyRule, AddPolicy whenNotEmptyRule) {
            this.whenEmptyRule = whenEmptyRule;
            this.whenNotEmptyRule = whenNotEmptyRule;
        }

        public boolean canAdd(Card card, Pile pile) {
            if (pile.isEmpty()) {
                return whenEmptyRule.canAdd(card, pile);
            } else {
                return whenNotEmptyRule.canAdd(card, pile);
            }
        }
    }

    public static final class SameDeck extends AddPolicy {

        private final Deck deck;

        public SameDeck(Deck deck) {
            this.deck = deck;
        }

        public boolean canAdd(Card card, Pile pile) {
            return card.getDeck() == deck;
        }
    }
}