package org.netbeans.modules.manifesterrorannotation;

import java.io.IOException;
import java.io.OutputStream;
import junit.framework.TestCase;
import javax.swing.text.StyledDocument;
import org.openide.cookies.SaveCookie;
import org.openide.filesystems.FileLock;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.loaders.DataObject;
import org.openide.text.Line;

/**
 *
 * @author Jaroslav Tulach
 */
public class ErrorAnnotationTest extends TestCase {
    ManifestEditor ed;
    DataObject obj;
    
    
    public ErrorAnnotationTest(String testName) {
        super(testName);
    }

    protected void setUp() throws Exception {
        FileObject fo = FileUtil.createData(
            FileUtil.createMemoryFileSystem().getRoot(), "x/manifest.mf"
        );
        FileUtil.setMIMEType("mf", "text/x-java-jar-manifest");
        obj = DataObject.find(fo);
        assertEquals(ManifestDataObject.class, obj.getClass());
        ed = (ManifestEditor)obj.getCookie(ManifestEditor.class);
    }

    protected void tearDown() throws Exception {
    }
    
    private void save() throws IOException {
        SaveCookie sc = (SaveCookie)obj.getCookie(SaveCookie.class);
        assertNotNull("Something to save", sc);
        sc.save();
        waitFinished();
    }

    private void waitFinished() {
        ed.findError.waitFinished();
    }
    
    public void testSimpleError() throws Exception {
        StyledDocument doc = ed.openDocument();
        doc.insertString(0, "ble\n\n\n\n\n\n\nName: x\n\n", null);
        
        save();
        Line[] lines = (Line[])ed.getLineSet().getLines().toArray(new Line[0]);
        int cnt = 0;
        for (int i = 0; i < lines.length; i++) {
            if (lines[i].getAnnotationCount() > 0) {
                assertEquals("Annotation is on the first line", 0, i);
                cnt++;
            }
        }
        if (cnt == 0) {
            fail("There should be an error annotation");
        }
    }
    public void testOkDocument() throws Exception {
        StyledDocument doc = ed.openDocument();
        doc.insertString(0, "V-X.A9: V-X.A9_U_ch_o\n" +
            "V-X.A9: V-X.A9/U_ch_o\n\n" +
            "\n" +
            "\n\nName: x\n" +
            "\nErr", null);
        
        waitFinished();
        
        int cnt = 0;
        for (int i = 0; i < 20; i++) {
            Line line;
            line = ed.getLineSet().getCurrent(i);
            if (line.getAnnotationCount() > 0) {
                try {
                    line = ed.getLineSet().getCurrent(i + 1);
                    fail("Error annotation shall be only on last line: of l" + i + " txt:\n" + line.getText());
                } catch (IndexOutOfBoundsException ex) {
                    break;
                }
            }
        }
    }

    public void testSimpleErrorOn3rd() throws Exception {
        StyledDocument doc = ed.openDocument();
        doc.insertString(0, "\n\n\nble\n\n\n\n\n\n\nName: x\n\n", null);

        save();
        Line[] lines = (Line[])ed.getLineSet().getLines().toArray(new Line[0]);
        int cnt = 0;
        for (int i = 0; i < lines.length; i++) {
            if (lines[i].getAnnotationCount() > 0) {
                assertEquals("Annotation is on the 3rd line", 3, i);
                cnt++;
            }
        }
        if (cnt == 0) {
            fail("There should be an error annotation");
        }
    }

    public void testSimpleErrorOnOpen() throws Exception {
        FileLock lock = obj.getPrimaryFile().lock();
        OutputStream os = obj.getPrimaryFile().getOutputStream(lock);
        os.write("\n\n\nble\n\n\n\n\n\n\nName: x\n\n".getBytes());
        os.close();
        lock.releaseLock();
        
        StyledDocument doc = ed.openDocument();
        ed.findError.waitFinished();

        Line[] lines = (Line[])ed.getLineSet().getLines().toArray(new Line[0]);
        int cnt = 0;
        for (int i = 0; i < lines.length; i++) {
            if (lines[i].getAnnotationCount() > 0) {
                assertEquals("Annotation is on the 3rd line", 3, i);
                cnt++;
            }
        }
        if (cnt == 0) {
            fail("There should be an error annotation");
        }
    }

    public void testErrorAfterEdits() throws Exception {
        StyledDocument doc = ed.openDocument();
        doc.insertString(0, "X: Ok\n\n\n\n\n\n\n\nName: x\n\n", null);

        save();
        Line[] lines = (Line[])ed.getLineSet().getLines().toArray(new Line[0]);
        for (int i = 0; i < lines.length; i++) {
            if (lines[i].getAnnotationCount() > 0) {
                fail("No error");
            }
        }
        
        doc.insertString(7, "x: y\n", null);

        waitFinished();
        
        doc.insertString(7, "ble\n", null);
        
        waitFinished();
        
        int cnt = 0;
        for (int i = 0; i < 20; i++) {
            Line line;
            try {
                line = ed.getLineSet().getCurrent(i);
            } catch (IndexOutOfBoundsException ex) {
                break;
            }
            if (line.getAnnotationCount() > 0) {
                assertEquals("Annotation is on some line", 2, i);
                cnt++;
            }
        }
        if (cnt == 0) {
            fail("There should be an error annotation");
        }
    }
}
