/*
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS HEADER.
 *
 * Copyright 1997-2008 Sun Microsystems, Inc. All rights reserved.
 *
 * The contents of this file are subject to the terms of either the GNU
 * General Public License Version 2 only ("GPL") or the Common
 * Development and Distribution License("CDDL") (collectively, the
 * "License"). You may not use this file except in compliance with the
 * License. You can obtain a copy of the License at
 * http://www.netbeans.org/cddl-gplv2.html
 * or nbbuild/licenses/CDDL-GPL-2-CP. See the License for the
 * specific language governing permissions and limitations under the
 * License.  When distributing the software, include this License Header
 * Notice in each file and include the License file at
 * nbbuild/licenses/CDDL-GPL-2-CP.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the GPL Version 2 section of the License file that
 * accompanied this code. If applicable, add the following below the
 * License Header, with the fields enclosed by brackets [] replaced by
 * your own identifying information:
 * "Portions Copyrighted [year] [name of copyright owner]"
 *
 * Contributor(s):
 *
 * The Original Software is NetBeans. The Initial Developer of the Original
 * Software is Sun Microsystems, Inc. Portions Copyright 1997-2008 Sun
 * Microsystems, Inc. All Rights Reserved.
 *
 * If you wish your version of this file to be governed by only the CDDL
 * or only the GPL Version 2, indicate your decision by adding
 * "[Contributor] elects to include this software in this distribution
 * under the [CDDL or GPL Version 2] license." If you do not indicate a
 * single choice of license, a recipient has the option to distribute
 * your version of this file under either the CDDL, the GPL Version 2 or
 * to extend the choice of license to its licensees as provided above.
 * However, if you add GPL Version 2 code and therefore, elected the GPL
 * Version 2 license, then the option applies only if the new code is
 * made subject to such option by the copyright holder.
 */
package org.netbeans.examples.nbtidyintegration;

import java.awt.Component;
import java.awt.Dialog;
import java.awt.Dimension;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.StringReader;
import javax.swing.JPanel;
import javax.swing.text.BadLocationException;
import javax.swing.text.StyledDocument;
import org.netbeans.api.diff.Diff;
import org.openide.DialogDescriptor;
import org.openide.DialogDisplayer;
import org.openide.ErrorManager;
import org.openide.NotifyDescriptor;
import org.openide.cookies.EditorCookie;
import org.openide.execution.NbProcessDescriptor;
import org.openide.loaders.DataObject;
import org.openide.nodes.Node;
import org.openide.util.NbBundle;
import org.openide.util.RequestProcessor;
import org.openide.util.Task;
import org.openide.windows.OutputWriter;
import org.netbeans.examples.nbtidyintegration.utils.PrintInputStreamRunnable;
import org.netbeans.examples.nbtidyintegration.utils.ReadOutputStreamRunnable;

/**
 * Do error checking on HTML files by calling tidy.
 *
 * Reads only error stream
 */
public class TidyConversionRunnable extends TidyRunnable {

    boolean preview = false;
    private static RequestProcessor processor = null;

    /** Initialize HTML-Tidy runnable.*/
    public TidyConversionRunnable(Node node, String commandLineArgs) {
        this(node, commandLineArgs, false);
    }

    /** Initialize HTML-Tidy runnable.*/
    public TidyConversionRunnable(Node node, String commandLineArgs, boolean preview) {
        super(new Node[]{node}, commandLineArgs);
        this.preview = preview;
    }

    /**
     * Perform external program with this data object.
     * @param executable Absolute path to the executable
     * @param writer Writer for the output tab
     * @param dataObject Data Object that is processed
     * @throws java.lang.InterruptedException
     * @throws java.io.IOException
     */
    @Override
    void processDataObject(final OutputWriter writer,
            final String executable, final DataObject dataObject)
            throws IOException {
        try {
            writer.println("Running: " + executable + " Arguments: " + getCommandLineArgs());
            StyledDocument doc = getDocument(dataObject);
            //TODO: if (doc==null), it's not open in the editor: use file mod instead?
            String content = doc.getText(0, doc.getLength());

            NbProcessDescriptor tidyProcessDesc =
                    new NbProcessDescriptor(executable, getCommandLineArgs());
            Process process = tidyProcessDesc.exec();

            if (processor == null) {
                processor = new RequestProcessor("ConversionRunnable", 1, true);
            }
            processor.post(new PrintInputStreamRunnable(
                    process.getOutputStream(), content));
            BufferedReader outReader = new BufferedReader(new InputStreamReader(process.getInputStream()), 8192);
            ReadOutputStreamRunnable outRunnable = new ReadOutputStreamRunnable(
                    outReader);

            Task outTask = processor.post(outRunnable);
            readOutput(writer, dataObject, process.getErrorStream());
            process.waitFor();
            outTask.waitFinished(10000);
            String result = outRunnable.getResult();
            try {
                outReader.close();
                process.getErrorStream().close();
            } catch (IOException ioe) {
                //Stream already closed, this is fine
                ErrorManager.getDefault().notify(ErrorManager.WARNING, ioe);
            }

            if (!preview || diff(content, result, dataObject.getPrimaryFile().getMIMEType())) {
                doc.remove(0, doc.getLength());
                doc.insertString(0, result, null);
            }
        } catch (BadLocationException e) {
            ErrorManager.getDefault().notify(ErrorManager.WARNING, e);
        } catch (InterruptedException e) {
            ErrorManager.getDefault().notify(ErrorManager.WARNING, e);
        }
        return;
    }

    public static StyledDocument getDocument(DataObject dataObject) {
        final EditorCookie edit = (EditorCookie) dataObject.getCookie(EditorCookie.class);
        if (edit == null) {
            ErrorManager.getDefault().log(ErrorManager.USER, "no editor cookie!");
            return null;
        }
        return edit.getDocument();
    }

    static boolean diff(String before, String after, String mime) {
        Diff diff = Diff.getDefault();
        JPanel panel = new JPanel();
        Component tp = null;
        StringReader bReader = new StringReader(before);
        StringReader bAfter = new StringReader(after);
        try {
            tp = diff.createDiff("DiffBefore", NbBundle.getMessage(TidyConversionRunnable.class,
                    "TitleDiffBefore"), bReader, "DiffAfter",
                    NbBundle.getMessage(TidyConversionRunnable.class,
                    "TitleDiffAfter"), bAfter, mime);
            tp.setMinimumSize(new Dimension(320, 600));
            panel.setMinimumSize(new Dimension(320, 600));

            panel.add(tp);
            DialogDescriptor d = new DialogDescriptor(tp,
                    NbBundle.getMessage(TidyConversionRunnable.class,
                    "TitleDiff")); // NOI18N
            d.setModal(true);
            d.setMessageType(NotifyDescriptor.PLAIN_MESSAGE);
            d.setOptionType(NotifyDescriptor.DEFAULT_OPTION);
            Dialog dlg = DialogDisplayer.getDefault().createDialog(d);
            dlg.pack();
            dlg.setSize(new Dimension(800, 600));
            dlg.setVisible(true);
            return (d.getValue() == NotifyDescriptor.OK_OPTION);
        } catch (IOException ioex) {
            ErrorManager.getDefault().notify(ioex);
            return false;
        }
    }
}

